import warnings

import inspect
import json
import logging
import uuid
from typing import Callable
from enum import Enum
from typing import List
import datetime

from .. import entities, assets, repositories, PlatformException

NODE_SIZE = (200, 87)

logger = logging.getLogger(name='dtlpy')


class PipelineConnectionPort:
    def __init__(self, node_id: str, port_id: str):
        self.node_id = node_id
        self.port_id = port_id

    @staticmethod
    def from_json(_json: dict):
        return PipelineConnectionPort(
            node_id=_json.get('nodeId', None),
            port_id=_json.get('portId', None),
        )

    def to_json(self):
        _json = {
            'nodeId': self.node_id,
            'portId': self.port_id,
        }
        return _json


class PipelineConnection:
    def __init__(self,
                 source: PipelineConnectionPort,
                 target: PipelineConnectionPort,
                 filters: entities.Filters,
                 action: str = None
                 ):
        """
        :param PipelineConnectionPort source: the source pipeline connection
        :param PipelineConnectionPort target: the target pipeline connection
        :param entities.Filters filters: condition for the connection between the nodes
        :param str action: the action that move the input when it happen
        """
        self.source = source
        self.target = target
        self.filters = filters
        self.action = action

    @staticmethod
    def from_json(_json: dict):
        condition = _json.get('condition', None)
        if condition:
            condition = json.loads(condition)
        return PipelineConnection(
            source=PipelineConnectionPort.from_json(_json=_json.get('src', None)),
            target=PipelineConnectionPort.from_json(_json=_json.get('tgt', None)),
            filters=condition,
            action=_json.get('action', None)
        )

    def to_json(self):
        _json = {
            'src': self.source.to_json(),
            'tgt': self.target.to_json(),
        }
        if self.action:
            _json['action'] = self.action
        if self.filters:
            if isinstance(self.filters, entities.Filters):
                filters = self.filters.prepare(query_only=True).get('filter', dict())
            else:
                filters = self.filters

            _json['condition'] = json.dumps(filters)
        return _json


class PipelineNodeIO:
    def __init__(self,
                 input_type: entities.PackageInputType,
                 name: str,
                 display_name: str,
                 port_id: str = None,
                 color: tuple = None,
                 port_percentage: int = None,
                 default_value=None,
                 variable_name: str = None,
                 actions: list = None,
                 description: str = None):
        """
        Pipeline Node

        :param entities.PackageInputType input_type: entities.PackageInputType of the input type of the pipeline
        :param str name: name of the input
        :param str display_name: of the input
        :param str port_id: port id
        :param tuple color: tuple the display the color
        :param int port_percentage: port percentage
        :param str action: the action that move the input when it happen
        :param default_value: default value of the input
        :param list actions: the actions list that move the input when it happen
        """
        self.port_id = port_id if port_id else str(uuid.uuid4())
        self.input_type = input_type
        self.name = name
        self.color = color
        self.display_name = display_name
        self.port_percentage = port_percentage
        self.default_value = default_value
        self.variable_name = variable_name
        self.description = description
        self.actions = actions

    @staticmethod
    def from_json(_json: dict):
        return PipelineNodeIO(
            port_id=_json.get('portId', None),
            input_type=_json.get('type', None),
            name=_json.get('name', None),
            color=_json.get('color', None),
            display_name=_json.get('displayName', None),
            port_percentage=_json.get('portPercentage', None),
            default_value=_json.get('defaultValue', None),
            variable_name=_json.get('variableName', None),
            actions=_json.get('actions', None),
            description=_json.get('description', None),
        )

    def to_json(self):
        _json = {
            'portId': self.port_id,
            'type': self.input_type,
            'name': self.name,
            'color': self.color,
            'displayName': self.display_name,
            'variableName': self.variable_name,
            'portPercentage': self.port_percentage,
        }

        if self.actions:
            _json['actions'] = self.actions
        if self.default_value:
            _json['defaultValue'] = self.default_value
        if self.description:
            _json['description'] = self.description
        return _json


class PipelineNodeType(str, Enum):
    TASK = 'task'
    CODE = 'code'
    FUNCTION = 'function'
    STORAGE = 'storage'
    ML = 'ml'


class PipelineNameSpace:
    def __init__(self, function_name, project_name, module_name=None, service_name=None, package_name=None):
        self.function_name = function_name
        self.project_name = project_name
        self.module_name = module_name
        self.service_name = service_name
        self.package_name = package_name

    def to_json(self):
        _json = {
            "functionName": self.function_name,
            "projectName": self.project_name
        }
        if self.module_name:
            _json['moduleName'] = self.module_name

        if self.service_name:
            _json['serviceName'] = self.service_name

        if self.package_name:
            _json['packageName'] = self.package_name
        return _json

    @staticmethod
    def from_json(_json: dict):
        return PipelineNameSpace(
            function_name=_json.get('functionName'),
            project_name=_json.get('projectName'),
            module_name=_json.get('moduleName', None),
            service_name=_json.get('serviceName', None),
            package_name=_json.get('packageName', None)
        )


class PipelineNode:
    def __init__(self,
                 name: str,
                 node_id: str,
                 outputs: list,
                 inputs: list,
                 node_type: PipelineNodeType,
                 namespace: PipelineNameSpace,
                 project_id: str,
                 metadata: dict = None,
                 config: dict = None,
                 position: tuple = (1, 1),
                 app_id: str = None,
                 dpk_name: str = None,
                 app_name: str = None,
                 ):
        """
        :param str name: node name
        :param str node_id: node id
        :param list outputs: list of PipelineNodeIO outputs
        :param list inputs: list of PipelineNodeIO inputs
        :param dict metadata: dict of the metadata of the node
        :param PipelineNodeType node_type: task, code, function
        :param PipelineNameSpace namespace: PipelineNameSpace of the node space
        :param str project_id: project id
        :param dict config: for the code node dict in format { package: {code : the_code}}
        :param tuple position: tuple of the node place
        :param str app_id: app id
        :param str dpk_name: dpk name
        :param str app_name: app name
        """
        self.name = name
        self.node_id = node_id
        self.outputs = outputs
        self.inputs = inputs
        self.metadata = metadata if metadata is not None else {}
        self.node_type = node_type
        self.namespace = namespace
        self.project_id = project_id
        self.config = config
        self.position = position
        self.app_id = app_id
        self.dpk_name = dpk_name
        self.app_name = app_name
        self._pipeline = None

    @property
    def position(self):
        position_tuple = (self.metadata['position']['x'],
                          self.metadata['position']['y'])
        return position_tuple

    @position.setter
    def position(self, position):
        self.metadata['position'] = \
            {
                "x": position[0] * 1.7 * NODE_SIZE[0] + NODE_SIZE[0] / 2,
                "y": position[1] * 1.5 * NODE_SIZE[1] + NODE_SIZE[1],
                "z": 0
            }

    def _default_io(self, actions: list = None) -> PipelineNodeIO:
        """
        Create a default item pipeline input

        :param str actions:  the action that move the input when it happen
        :return PipelineNodeIO: the default item PipelineNodeIO
        """
        default_io = PipelineNodeIO(port_id=str(uuid.uuid4()),
                                    input_type=entities.PackageInputType.ITEM,
                                    name='item',
                                    color=None,
                                    display_name=actions[0] if actions else 'item',
                                    actions=actions)
        return default_io

    @staticmethod
    def from_json(_json: dict):
        inputs = [PipelineNodeIO.from_json(_json=i_input) for i_input in _json.get('inputs', list())]
        outputs = [PipelineNodeIO.from_json(_json=i_output) for i_output in _json.get('outputs', list())]
        namespace = PipelineNameSpace.from_json(_json.get('namespace', {}))
        metadata = _json.get('metadata', {})
        position = ((metadata['position']['x'] - NODE_SIZE[0] / 2) / (1.7 * NODE_SIZE[0]),
                    (metadata['position']['y'] - NODE_SIZE[1]) / (1.5 * NODE_SIZE[1]))
        return PipelineNode(
            name=_json.get('name', None),
            node_id=_json.get('id', None),
            outputs=outputs,
            inputs=inputs,
            metadata=metadata,
            node_type=_json.get('type', None),
            namespace=namespace,
            project_id=_json.get('projectId', None),
            config=_json.get('config', None),
            position=position,
            app_id=_json.get('appId', None),
            dpk_name=_json.get('dpkName', None),
            app_name=_json.get('appName', None),
        )

    def to_json(self):
        _json = {
            'name': self.name,
            'id': self.node_id,
            'outputs': [_io.to_json() for _io in self.outputs],
            'inputs': [_io.to_json() for _io in self.inputs],
            'metadata': self.metadata,
            'type': self.node_type,
            'namespace': self.namespace.to_json(),
            'projectId': self.project_id,
            'dpkName': self.dpk_name,
            'appName': self.app_name,
        }
        if self.config is not None:
            _json['config'] = self.config
        if self.app_id is not None:
            _json['appId'] = self.app_id
        return _json

    def is_root(self):
        if self._pipeline is not None:
            for node in self._pipeline.start_nodes:
                if self.node_id == node.get('nodeId', None) and node.get('type', None) == 'root':
                    return True
        return False

    def _build_connection(self,
                          node,
                          source_port: PipelineNodeIO = None,
                          target_port: PipelineNodeIO = None,
                          filters: entities.Filters = None,
                          action: str = None) -> PipelineConnection:
        """
        Build connection between the current node and the target node use the given ports

        :param PipelineNode node: the node to connect to it
        :param PipelineNodeIO source_port: the source PipelineNodeIO input port
        :param PipelineNodeIO target_port: the target PipelineNodeIO output port
        :param entities.Filters filters: condition for the connection between the nodes
        :param str action:  the action that move the input when it happen
        :return: the connection between the nodes
        """
        if source_port is None and self.outputs:
            source_port = self.outputs[0]

        if target_port is None and node.inputs:
            target_port = node.inputs[0]

        if node.is_root():
            self._pipeline.set_start_node(self)

        source_connection = PipelineConnectionPort(node_id=self.node_id, port_id=source_port.port_id)
        target_connection = PipelineConnectionPort(node_id=node.node_id, port_id=target_port.port_id)
        if action is None and source_port.actions is not None and source_port.actions != []:
            action = source_port.actions[0]
        connection = PipelineConnection(source=source_connection, target=target_connection, filters=filters,
                                        action=action)
        return connection

    def connect(self,
                node,
                source_port: PipelineNodeIO = None,
                target_port: PipelineNodeIO = None,
                filters=None,
                action: str = None):
        """
        Build connection between the current node and the target node use the given ports

        :param PipelineNode node: the node to connect to it
        :param PipelineNodeIO source_port: the source PipelineNodeIO input port
        :param PipelineNodeIO target_port: the target PipelineNodeIO output port
        :param entities.Filters filters: condition for the connection between the nodes
        :param str action:  the action that move the input when it happen
        :return: the connected node
        """
        if self._pipeline is None:
            raise Exception("must add the node to the pipeline first, e.g pipeline.nodes.add(node)")
        connection = self._build_connection(node=node,
                                            source_port=source_port,
                                            target_port=target_port,
                                            filters=filters,
                                            action=action)
        self._pipeline.connections.append(connection)
        self._pipeline.nodes.add(node)
        return node

    def disconnect(self,
                   node,
                   source_port: PipelineNodeIO = None,
                   target_port: PipelineNodeIO = None) -> bool:
        """
        remove connection between the current node and the target node use the given ports

        :param PipelineNode node: the node to connect to it
        :param PipelineNodeIO source_port: the source PipelineNodeIO input port
        :param PipelineNodeIO target_port: the target PipelineNodeIO output port
        :return: true if success and false if not
        """
        if self._pipeline is None:
            raise Exception("must add the node to the pipeline first, e.g pipeline.nodes.add(node)")
        connection = self._build_connection(node=node,
                                            source_port=source_port,
                                            target_port=target_port,
                                            filters=None)

        current_connection = connection.to_json()
        if 'condition' in current_connection:
            current_connection = current_connection.pop('condition')

        for connection_index in range(len(self._pipeline.connections)):
            pipeline_connection = self._pipeline.connections[connection_index].to_json()
            if 'condition' in pipeline_connection:
                pipeline_connection = pipeline_connection.pop('condition')

            if current_connection == pipeline_connection:
                self._pipeline.connections.pop(connection_index)
                return True
        logger.warning('do not found a connection')
        return False

    def add_trigger(self,
                    trigger_type: entities.TriggerType = entities.TriggerType.EVENT,
                    filters=None,
                    resource: entities.TriggerResource = entities.TriggerResource.ITEM,
                    actions: entities.TriggerAction = entities.TriggerAction.CREATED,
                    execution_mode: entities.TriggerExecutionMode = entities.TriggerExecutionMode.ONCE,
                    cron: str = None,
                    ):
        """
        Create a Trigger. Can create two types: a cron trigger or an event trigger.
        Inputs are different for each type

        Inputs for all types:

        :param trigger_type: can be cron or event. use enum dl.TriggerType for the full list

        Inputs for event trigger:
        :param filters: optional - Item/Annotation metadata filters, default = none
        :param resource: optional - Dataset/Item/Annotation/ItemStatus, default = Item
        :param actions: optional - Created/Updated/Deleted, default = create
        :param execution_mode: how many time trigger should be activate. default is "Once". enum dl.TriggerExecutionMode

        Inputs for cron trigger:
        :param str cron: cron spec specifying when it should run. more information: https://en.wikipedia.org/wiki/Cron

        :return: Trigger entity
        """
        if self._pipeline is None:
            raise Exception("must add the node to the pipeline first, e.g pipeline.nodes.add(node)")

        if not isinstance(actions, list):
            actions = [actions]

        if filters is None:
            filters = {}
        else:
            filters = json.dumps(filters.prepare(query_only=True).get('filter', dict()))

        if trigger_type == entities.TriggerType.EVENT:
            spec = {
                'filter': filters,
                'resource': resource,
                'executionMode': execution_mode,
                'actions': actions
            }
        elif trigger_type == entities.TriggerType.CRON:
            spec = {
                'cron': cron,
            }
        else:
            raise ValueError('Unknown trigger type: "{}". Use dl.TriggerType for known types'.format(trigger_type))

        trigger = {
            "type": trigger_type,
            "spec": spec
        }

        set_trigger = False
        for pipe_node in self._pipeline.start_nodes:
            if pipe_node['nodeId'] == self.node_id:
                set_trigger = True
                pipe_node['trigger'] = trigger

        if not set_trigger:
            self._pipeline.start_nodes.append(
                {
                    "nodeId": self.node_id,
                    "type": "trigger",
                    'trigger': trigger
                }
            )


class CodeNode(PipelineNode):
    def __init__(self,
                 name: str,
                 project_id: str,
                 project_name: str,
                 method: Callable,
                 outputs: List[PipelineNodeIO] = None,
                 inputs: List[PipelineNodeIO] = None,
                 position: tuple = (1, 1),
                 ):
        """
        :param str name: node name
        :param str project_id: project id
        :param str project_name: project name
        :param Callable method: function to deploy
        :param list outputs: list of PipelineNodeIO outputs
        :param list inputs: list of PipelineNodeIO inputs
        :param tuple position: tuple of the node place
        """
        if inputs is None:
            inputs = [self._default_io()]
        if outputs is None:
            outputs = [self._default_io()]

        if method is None or not isinstance(method, Callable):
            raise Exception('must provide a function as input')
        else:
            function_code = self._build_code_from_func(method)
            function_name = method.__name__

        super().__init__(name=name,
                         node_id=str(uuid.uuid4()),
                         outputs=outputs,
                         inputs=inputs,
                         metadata={},
                         node_type=PipelineNodeType.CODE,
                         namespace=PipelineNameSpace(function_name=function_name, project_name=project_name),
                         project_id=project_id,
                         position=position)

        self.config = {
            "package":
                {
                    "code": function_code,
                    "name": function_name,
                    "type": "code"
                }
        }

    def _build_code_from_func(self, func: Callable) -> str:
        """
        Build a code format from the given function

        :param Callable func: function to deploy
        :return: a string the display the code with the package format
        """
        with open(assets.paths.PARTIAL_MAIN_FILEPATH, 'r') as f:
            main_string = f.read()
        lines = inspect.getsourcelines(func)

        tabs_diff = lines[0][0].count('    ') - 1
        for line_index in range(len(lines[0])):
            line_tabs = lines[0][line_index].count('    ') - tabs_diff
            lines[0][line_index] = ('    ' * line_tabs) + lines[0][line_index].strip() + '\n'

        method_func_string = "".join(lines[0])

        code = '{}\n{}\n    @staticmethod\n{}'.format('', main_string,
                                                      method_func_string)
        return code

    @staticmethod
    def from_json(_json: dict):
        parent = PipelineNode.from_json(_json)
        parent.__class__ = CodeNode
        return parent


class TaskNode(PipelineNode):
    def __init__(self,
                 name: str,
                 project_id: str,
                 dataset_id: str,
                 recipe_title: str,
                 recipe_id: str,
                 task_owner: str,
                 workload: List[entities.WorkloadUnit],
                 task_type: str = 'annotation',
                 position: tuple = (1, 1),
                 actions: list = None,
                 repeatable: bool = True,
                 batch_size=None,
                 max_batch_workload=None,
                 priority=entities.TaskPriority.MEDIUM,
                 due_date=None,
                 consensus_task_type=None,
                 consensus_percentage=None,
                 consensus_assignees=None,
                 groups=None
                 ):
        """
        :param str name: node name
        :param str project_id: project id
        :param str dataset_id: dataset id
        :param str recipe_title: recipe title
        :param str recipe_id: recipe id
        :param str task_owner: email of task owner
        :param List[WorkloadUnit] workload: list of WorkloadUnit
        :param str task_type: 'annotation' or 'qa'
        :param tuple position: tuple of the node place
        :param list actions: list of task actions
        :param bool repeatable: can repeat in the item
        :param int groups: groups to assign the task to
        :param int batch_size: Pulling batch size (items) . Restrictions - Min 3, max 100 - for create pulling task
        :param int max_batch_workload: Max items in assignment . Restrictions - Min batchSize + 2 , max batchSize * 2 - for create pulling task
        :param entities.TaskPriority priority: priority of the task options in entities.TaskPriority
        :param float due_date: date by which the task should be finished; for example, due_date = datetime.datetime(day= 1, month= 1, year= 2029).timestamp()
        :param entities.ConsensusTaskType consensus_task_type: consensus_task_type of the task options in entities.ConsensusTaskType
        :param int consensus_percentage: percentage of items to be copied to multiple annotators (consensus items)
        :param int consensus_assignees: the number of different annotators per item (number of copies per item)
        """
        if actions is None or actions == []:
            actions = []
            if task_type == 'qa':
                if 'approve' not in actions:
                    actions.append('approve')
            else:
                if 'complete' not in actions:
                    actions.append('complete')
            actions.append('discard')
        else:
            logger.warning(
                "The 'actions' field was updated to override the system default actions for task (complete/approve, discard) if provided, due to a bug fix.")

        inputs = [self._default_io()]

        outputs = [self._default_io(actions=actions)]

        if groups is not None:
            if not isinstance(groups, list) or not all(isinstance(group, str) for group in groups):
                raise ValueError('groups must be a list of strings')

        super().__init__(name=name,
                         node_id=str(uuid.uuid4()),
                         outputs=outputs,
                         inputs=inputs,
                         metadata=dict(),
                         node_type=PipelineNodeType.TASK,
                         namespace=PipelineNameSpace(function_name="move_to_task",
                                                     project_name="DataloopTasks",
                                                     service_name="pipeline-utils"),
                         project_id=project_id,
                         position=position)

        self.dataset_id = dataset_id
        self.recipe_title = recipe_title
        self.recipe_id = recipe_id
        self.task_owner = task_owner
        self.task_type = task_type
        if not isinstance(workload, list):
            workload = [workload]
        self.workload = workload
        self.repeatable = repeatable
        if max_batch_workload:
            self.max_batch_workload = max_batch_workload
        if batch_size:
            self.batch_size = batch_size
        if consensus_task_type:
            self.consensus_task_type = consensus_task_type
        if consensus_percentage:
            self.consensus_percentage = consensus_percentage
        if consensus_assignees:
            self.consensus_assignees = consensus_assignees
        self.priority = priority
        if due_date is None:
            due_date = (datetime.datetime.now() + datetime.timedelta(days=7)).timestamp() * 1000
        self.due_date = due_date
        self.groups = groups

    @property
    def dataset_id(self):
        return self.metadata['datasetId']

    @dataset_id.setter
    def dataset_id(self, dataset_id: str):
        if not isinstance(dataset_id, str):
            raise PlatformException('400', 'Param dataset_id must be of type string')
        self.metadata['datasetId'] = dataset_id

    @property
    def groups(self):
        return self.metadata.get('groups')

    @groups.setter
    def groups(self, groups: List[str]):
        if groups is not None:
            self.metadata['groups'] = groups

    @property
    def repeatable(self):
        return self.metadata['repeatable']

    @repeatable.setter
    def repeatable(self, repeatable: bool):
        if not isinstance(repeatable, bool):
            raise PlatformException('400', 'Param repeatable must be of type bool')
        self.metadata['repeatable'] = repeatable

    @property
    def recipe_title(self):
        return self.metadata['recipeTitle']

    @recipe_title.setter
    def recipe_title(self, recipe_title: str):
        if not isinstance(recipe_title, str):
            raise PlatformException('400', 'Param recipe_title must be of type string')
        self.metadata['recipeTitle'] = recipe_title

    @property
    def recipe_id(self):
        return self.metadata['recipeId']

    @recipe_id.setter
    def recipe_id(self, recipe_id: str):
        if not isinstance(recipe_id, str):
            raise PlatformException('400', 'Param recipe_id must be of type string')
        self.metadata['recipeId'] = recipe_id

    @property
    def task_owner(self):
        return self.metadata['taskOwner']

    @task_owner.setter
    def task_owner(self, task_owner: str):
        if not isinstance(task_owner, str):
            raise PlatformException('400', 'Param task_owner must be of type string')
        self.metadata['taskOwner'] = task_owner

    @property
    def task_type(self):
        return self.metadata['taskType']

    @task_type.setter
    def task_type(self, task_type: str):
        if not isinstance(task_type, str):
            raise PlatformException('400', 'Param task_type must be of type string')
        self.metadata['taskType'] = task_type

    @property
    def workload(self):
        return self.metadata['workload']

    @workload.setter
    def workload(self, workload: list):
        if not isinstance(workload, list):
            workload = [workload]
        self.metadata['workload'] = [val.to_json() for val in workload]

    @property
    def batch_size(self):
        return self.metadata['batchSize']

    @batch_size.setter
    def batch_size(self, batch_size: int):
        if not isinstance(batch_size, int):
            raise PlatformException('400', 'Param batch_size must be of type int')
        self.metadata['batchSize'] = batch_size

    @property
    def max_batch_workload(self):
        return self.metadata['maxBatchWorkload']

    @max_batch_workload.setter
    def max_batch_workload(self, max_batch_workload: int):
        if not isinstance(max_batch_workload, int):
            raise PlatformException('400', 'Param max_batch_workload must be of type int')
        self.metadata['maxBatchWorkload'] = max_batch_workload

    @property
    def consensus_task_type(self):
        return self.metadata['consensusTaskType']

    @consensus_task_type.setter
    def consensus_task_type(self, consensus_task_type: entities.ConsensusTaskType):
        if not isinstance(consensus_task_type, str) and not isinstance(consensus_task_type, entities.ConsensusTaskType):
            raise PlatformException('400', 'Param consensus_task_type must be of type entities.ConsensusTaskType')
        self.metadata['consensusTaskType'] = consensus_task_type

    @property
    def consensus_percentage(self):
        return self.metadata['consensusPercentage']

    @consensus_percentage.setter
    def consensus_percentage(self, consensus_percentage: int):
        if not isinstance(consensus_percentage, int):
            raise PlatformException('400', 'Param consensus_percentage must be of type int')
        self.metadata['consensusPercentage'] = consensus_percentage

    @property
    def consensus_assignees(self):
        return self.metadata['consensusAssignees']

    @consensus_assignees.setter
    def consensus_assignees(self, consensus_assignees: int):
        if not isinstance(consensus_assignees, int):
            raise PlatformException('400', 'Param consensus_assignees must be of type int')
        self.metadata['consensusAssignees'] = consensus_assignees

    @property
    def priority(self):
        return self.metadata['priority']

    @priority.setter
    def priority(self, priority: entities.TaskPriority):
        if not isinstance(priority, int) and not isinstance(priority, entities.TaskPriority):
            raise PlatformException('400', 'Param priority must be of type entities.TaskPriority')
        self.metadata['priority'] = priority

    @property
    def due_date(self):
        return self.metadata['dueDate']

    @due_date.setter
    def due_date(self, due_date: float):
        if not isinstance(due_date, float) and not isinstance(due_date, int):
            raise PlatformException('400', 'Param due_date must be of type float or int')
        self.metadata['dueDate'] = due_date

    @staticmethod
    def from_json(_json: dict):
        parent = PipelineNode.from_json(_json)
        parent.__class__ = TaskNode
        return parent


class FunctionNode(PipelineNode):
    def __init__(self,
                 name: str,
                 service: entities.Service,
                 function_name,
                 position: tuple = (1, 1),
                 project_id=None,
                 project_name=None
                 ):
        """
        :param str name: node name
        :param entities.Service service: service to deploy
        :param str function_name: function name
        :param tuple position: tuple of the node place
        """
        self.service = service

        if project_id is None:
            project_id = service.project_id
        if project_id != service.project_id:
            logger.warning("the project id that provide different from the service project id")

        if project_name is None:
            try:
                project = repositories.Projects(client_api=self.service._client_api).get(project_id=project_id,
                                                                                         log_error=False)
                project_name = project.name
            except:
                logger.warning(
                    'Service project not found using DataloopTasks project.'
                    ' If this is incorrect please provide project_name param.')
                project_name = 'DataloopTasks'
        inputs = []
        outputs = []
        package = self.service.package
        modules = []
        if isinstance(package, entities.Package):
            modules = package.modules
        elif isinstance(package, entities.Dpk):
            modules = package.components.modules
        for model in modules:
            if model.name == self.service.module_name:
                for func in model.functions:
                    if func.name == function_name:
                        inputs = self._convert_from_function_io_to_pipeline_io(func.inputs)
                        outputs = self._convert_from_function_io_to_pipeline_io(func.outputs)

        namespace = PipelineNameSpace(
            function_name=function_name,
            service_name=self.service.name,
            module_name=self.service.module_name,
            package_name=self.service.package.name,
            project_name=project_name
        )
        super().__init__(name=name,
                         node_id=str(uuid.uuid4()),
                         outputs=outputs,
                         inputs=inputs,
                         metadata={},
                         node_type=PipelineNodeType.FUNCTION,
                         namespace=namespace,
                         project_id=service.project_id,
                         position=position)

    def _convert_from_function_io_to_pipeline_io(self, function_io: List[entities.FunctionIO]) -> List[PipelineNodeIO]:
        """
        Get a list of FunctionIO and convert them to PipelineIO
        :param List[entities.FunctionIO] function_io: list of functionIO
        :return: list of PipelineIO
        """
        pipeline_io = []
        for single_input in function_io:
            pipeline_io.append(
                PipelineNodeIO(port_id=str(uuid.uuid4()),
                               input_type=single_input.type,
                               name=single_input.name,
                               color=None,
                               display_name=single_input.name,
                               default_value=single_input.value,
                               actions=single_input.actions if single_input.actions is not None else []))
        return pipeline_io

    @staticmethod
    def from_json(_json: dict):
        parent = PipelineNode.from_json(_json)
        parent.__class__ = FunctionNode
        return parent


class DatasetNode(PipelineNode):
    def __init__(self,
                 name: str,
                 project_id: str,
                 dataset_id: str,
                 dataset_folder: str = None,
                 load_existing_data: bool = False,
                 data_filters: entities.Filters = None,
                 position: tuple = (1, 1)):
        """
        :param str name: node name
        :param str project_id: project id
        :param str dataset_id: dataset id
        :param str dataset_folder: folder in dataset to work in it
        :param bool load_existing_data: optional - enable to automatically load existing data into the
                                        pipeline (executions) upon activation, based on the defined dataset,
                                        folder, and data_filters.
        :param entities.Filters data_filters: optional - filters entity or a dictionary containing filters parameters.
                                              Use to filter the data items to be loaded when load_existing_data
                                              is enabled.
        :param tuple position: tuple of the node place
        """
        inputs = [self._default_io()]
        outputs = [self._default_io()]
        super().__init__(name=name,
                         node_id=str(uuid.uuid4()),
                         outputs=outputs,
                         inputs=inputs,
                         metadata={},
                         node_type=PipelineNodeType.STORAGE,
                         namespace=PipelineNameSpace(function_name="dataset_handler",
                                                     project_name="DataloopTasks",
                                                     service_name="pipeline-utils"),
                         project_id=project_id,
                         position=position)
        self.dataset_id = dataset_id
        self.dataset_folder = dataset_folder
        self.load_existing_data = load_existing_data
        self.data_filters = data_filters

    @property
    def dataset_id(self):
        return self.metadata['datasetId']

    @dataset_id.setter
    def dataset_id(self, dataset_id: str):
        self.metadata['datasetId'] = dataset_id

    @property
    def dataset_folder(self):
        return self.metadata.get('dir', None)

    @dataset_folder.setter
    def dataset_folder(self, dataset_folder: str):
        if dataset_folder is not None:
            if not dataset_folder.startswith("/"):
                dataset_folder = '/' + dataset_folder
            self.metadata['dir'] = dataset_folder

    @property
    def load_existing_data(self):
        return self.metadata.get('triggerToPipeline', {}).get('active', False)

    @load_existing_data.setter
    def load_existing_data(self, load_existing_data: bool):
        if load_existing_data:
            self.metadata.setdefault('triggerToPipeline', {})['active'] = True
        else:
            self.metadata.pop('triggerToPipeline', None)

    @property
    def data_filters(self):
        data_filters = self.metadata.get('triggerToPipeline', {}).get('filter', None)
        if data_filters:
            data_filters = entities.Filters(custom_filter=json.loads(data_filters))
        return data_filters

    @data_filters.setter
    def data_filters(self, data_filters: entities.Filters):
        if data_filters is None:
            filters = None
        else:
            filters = json.dumps(data_filters.prepare(query_only=True).get('filter'))
        self.metadata.setdefault('triggerToPipeline', {})['filter'] = filters

    @staticmethod
    def from_json(_json: dict):
        parent = PipelineNode.from_json(_json)
        parent.__class__ = DatasetNode
        return parent



import datetime
import inspect
import logging
import json
import os
import tempfile
import time
from typing import Union, List, Callable
from .. import miscellaneous, exceptions, entities, repositories, assets, ApiClient, _api_reference
from ..__version__ import version as __version__

logger = logging.getLogger(name='dtlpy')
FUNCTION_END_LINE = '[Done] Executing function.'
MAX_WAIT_TIME = 8


class Services:
    """
    Services Repository

    The Services class allows the user to manage services and their properties. Services are created from the packages users create.
    See our documentation for more information about `services <https://developers.dataloop.ai/tutorials/faas/advance/chapter/>`_.
    """

    def __init__(self,
                 client_api: ApiClient,
                 project: entities.Project = None,
                 package: Union[entities.Package, entities.Dpk] = None,
                 project_id=None,
                 model_id=None,
                 model: entities.Model = None):
        self._client_api = client_api
        self._package = package
        self._project = project
        self._model = model
        if model_id is None:
            if model is not None:
                model_id = model.id
        self._model_id = model_id
        if project_id is None:
            if project is not None:
                project_id = project.id
        self._project_id = project_id
        self._settings = repositories.Settings(project=project, client_api=client_api)

    ############
    # entities #
    ############
    @property
    def package(self) -> entities.Package:
        if self._package is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Cannot perform action WITHOUT package entity in services repository. Please set a package')
        assert (isinstance(self._package, entities.Package) or isinstance(self._package, entities.Dpk))
        return self._package

    @package.setter
    def package(self, package: Union[entities.Package, entities.Dpk]):
        if not isinstance(package, entities.Package) and not isinstance(package, entities.Dpk):
            raise ValueError('Must input a valid package entity')
        self._package = package

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            # try to get from package
            if self._package is not None:
                self._project = self._package._project

        if self._project is None:
            # try to get checked out project
            project = self._client_api.state_io.get('project')
            if project is not None:
                self._project = entities.Project.from_json(_json=project, client_api=self._client_api)
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Cannot perform action WITHOUT Project entity in services repository. Please set a project')
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project
        self._project_id = project.id

    @property
    def model(self) -> entities.Model:
        if self._model is None:
            if self._model_id is not None:
                self._model = self.project.models.get(model_id=self._model_id)
            else:
                raise exceptions.PlatformException(
                    error='2001',
                    message='Cannot perform action WITHOUT model entity in services repository. Please set a model')
        assert isinstance(self._model, entities.Model)
        return self._model

    @model.setter
    def model(self, model: entities.Model):
        if not isinstance(model, entities.Model):
            raise ValueError('Must input a valid model entity')
        self._model = model
        self._model_id = model.id

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/services".format(self.project.id))

    def open_in_web(self,
                    service: entities.Service = None,
                    service_id: str = None,
                    service_name: str = None):
        """
        Open the service in web platform

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param str service_name: service name
        :param str service_id: service id
        :param dtlpy.entities.service.Service service: service entity

        **Example**:

        .. code-block:: python

            package.services.open_in_web(service_id='service_id')
        """
        if service_name is not None:
            service = self.get(service_name=service_name)
        if service is not None:
            service.open_in_web()
        elif service_id is not None:
            self._client_api._open_in_web(url=self.platform_url + '/' + str(service_id) + '/main')
        else:
            self._client_api._open_in_web(url=self.platform_url)

    def __get_from_cache(self) -> entities.Service:
        service = self._client_api.state_io.get('service')
        if service is not None:
            service = entities.Service.from_json(_json=service,
                                                 client_api=self._client_api,
                                                 project=self._project,
                                                 package=self._package)
        return service

    def checkout(self,
                 service: entities.Service = None,
                 service_name: str = None,
                 service_id: str = None):
        """
        Checkout (switch) to a service.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param dtlpy.entities.service.Service service: Service entity
        :param str service_name: service name
        :param str service_id: service id

        **Example**:

        .. code-block:: python

            package.services.checkout(service_id='service_id')
        """
        if service is None:
            service = self.get(service_name=service_name, service_id=service_id)
        self._client_api.state_io.put('service', service.to_json())
        logger.info('Checked out to service {}'.format(service.name))

    ###########
    # methods #
    ###########
    @_api_reference.add(path='/services/{id}/revisions', method='get')
    def revisions(self,
                  service: entities.Service = None,
                  service_id: str = None):
        """
        Get service revisions history.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        You must provide at leats ONE of the following params: service, service_id

        :param dtlpy.entities.service.Service service: Service entity
        :param str service_id: service id

        **Example**:

        .. code-block:: python

            service_revision = package.services.revisions(service_id='service_id')
        """
        if service is None and service_id is None:
            raise exceptions.PlatformException(
                error='400',
                message='must provide an identifier in inputs: "service" or "service_id"')
        if service is not None:
            service_id = service.id

        success, response = self._client_api.gen_request(
            req_type="get",
            path="/services/{}/revisions".format(service_id))
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @_api_reference.add(path='/services/{id}', method='get')
    def get(self,
            service_name=None,
            service_id=None,
            checkout=False,
            fetch=None
            ) -> entities.Service:
        """
        Get service to use in your code.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param str service_name: optional - search by name
        :param str service_id: optional - search by id
        :param bool checkout: if true, checkout (switch) to service
        :param fetch: optional - fetch entity from platform, default taken from cookie
        :return: Service object
        :rtype: dtlpy.entities.service.Service

        **Example**:

        .. code-block:: python

            service = package.services.get(service_id='service_id')
        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if service_name is None and service_id is None:
            service = self.__get_from_cache()
            if service is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='No checked-out Service was found, must checkout or provide an identifier in inputs')

        elif fetch:
            if service_id is not None:
                success, response = self._client_api.gen_request(
                    req_type="get",
                    path="/services/{}".format(service_id)
                )
                if not success:
                    raise exceptions.PlatformException(response)
                service = entities.Service.from_json(client_api=self._client_api,
                                                     _json=response.json(),
                                                     package=self._package,
                                                     project=self._project)
                # verify input service name is same as the given id
                if service_name is not None and service.name != service_name:
                    logger.warning(
                        "Mismatch found in services.get: service_name is different then service.name:"
                        " {!r} != {!r}".format(
                            service_name,
                            service.name))
            elif service_name is not None:
                filters = entities.Filters(resource=entities.FiltersResource.SERVICE,
                                           field='name',
                                           values=service_name,
                                           use_defaults=False)
                if self._project_id is not None:
                    filters.add(field='projectId', values=self._project_id)
                if self._package is not None:
                    filters.add(field='packageId', values=self._package.id)
                services = self.list(filters=filters)
                if services.items_count > 1:
                    raise exceptions.PlatformException('404', 'More than one service with same name. '
                                                              'Please get services from package/project entity')
                elif services.items_count == 0:
                    raise exceptions.PlatformException('404', 'Service not found: {}.'.format(service_name))
                service = services.items[0]
            else:
                raise exceptions.PlatformException(
                    error='400',
                    message='No checked-out Service was found, must checkout or provide an identifier in inputs')
        else:
            service = entities.Service.from_json(_json={'id': service_id,
                                                        'name': service_name},
                                                 client_api=self._client_api,
                                                 project=self._project,
                                                 package=self._package,
                                                 is_fetched=False)

        assert isinstance(service, entities.Service)
        if checkout:
            self.checkout(service=service)
        return service

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Service]:
        jobs = [None for _ in range(len(response_items))]
        pool = self._client_api.thread_pools(pool_name='entity.create')

        # return triggers list
        for i_service, service in enumerate(response_items):
            jobs[i_service] = pool.submit(entities.Service._protected_from_json,
                                          **{'client_api': self._client_api,
                                             '_json': service,
                                             'package': self._package,
                                             'project': self._project})

        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])

    def _list(self, filters: entities.Filters):
        url = '/query/faas'
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    @_api_reference.add(path='/query/faas', method='post')
    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List all services (services can be listed for a package or for a project).

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            services = package.services.list()
        """
        # default filters
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.SERVICE)
        # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))
        if filters.resource != entities.FiltersResource.SERVICE:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.SERVICE. Got: {!r}'.format(filters.resource))
        if self._project_id is not None:
            filters.add(field='projectId', values=self._project_id)
        if self._model_id is not None:
            filters.add(field='metadata.ml.modelId', values=self._model_id)
        if self._package is not None:
            filters.add(field='packageId', values=self._package.id)

        # assert type filters
        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException('400', 'Unknown filters type')

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    @_api_reference.add(path='/services/{id}/status', method='post')
    def status(self, service_name=None, service_id=None):
        """
        Get service status.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        You must provide at least ONE of the following params: service_id, service_name

        :param str service_name: service name
        :param str service_id: service id
        :return: status json
        :rtype: dict

        **Example**:

        .. code-block:: python

            status_json = package.services.status(service_id='service_id')
        """
        if service_id is None:
            if service_name is None:
                raise exceptions.PlatformException(error='400',
                                                   message='must input "service_name" or "service_id" to get status')
            service = self.get(service_name=service_name)
            service_id = service.id
        # request
        success, response = self._client_api.gen_request(req_type="get",
                                                         path="/services/{}/status".format(service_id))
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @_api_reference.add(path='/services/{id}/stop', method='post')
    def pause(self,
              service_name: str = None,
              service_id: str = None,
              force: bool = False):
        """
        Pause service.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        You must provide at least ONE of the following params: service_id, service_name

        :param str service_name: service name
        :param str service_id: service id
        :param bool force: optional - terminate old replicas immediately
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = package.services.pause(service_id='service_id')
        """
        if service_id is None:
            if service_name is None:
                raise exceptions.PlatformException(error='400',
                                                   message='must input "service_name" or "service_id" to pause service')
            service = self.get(service_name=service_name)
            service_id = service.id
        # request
        url = "/services/{}/stop".format(service_id)
        if force:
            url = '{}?force=true'
        success, response = self._client_api.gen_request(req_type="post",
                                                         path=url)
        if not success:
            raise exceptions.PlatformException(response)
        return success

    def _notify(
            self,
            service_id: str,
            message: str,
            name: str,
            action: str = 'created',
            support: str = None,
            docs: str = None,
            agent_info: dict = None
    ):
        url = "/services/{}/notify".format(service_id)
        payload = {
            'action': action,
            'message': message,
            'notificationName': name
        }
        if agent_info is not None:
            payload['agentInfo'] = agent_info

        if support:
            payload['support'] = support

        if docs:
            payload['docs'] = docs

        success, response = self._client_api.gen_request(
            req_type="post",
            path=url,
            json_req=payload
        )
        if not success:
            raise exceptions.PlatformException(response)
        return success

    @_api_reference.add(path='/services/{id}/resume', method='post')
    def resume(self,
               service_name: str = None,
               service_id: str = None,
               force: bool = False):
        """
        Resume service.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        You must provide at least ONE of the following params: service_id, service_name.

        :param str service_name: service name
        :param str service_id: service id
        :param bool force: optional - terminate old replicas immediately
        :return: json of the service
        :rtype: dict

        **Example**:

        .. code-block:: python

            service_json = package.services.resume(service_id='service_id')
        """
        if service_id is None:
            if service_name is None:
                raise exceptions.PlatformException(error='400',
                                                   message='must input "service_name" or "service_id" to resume')
            service = self.get(service_name=service_name)
            service_id = service.id
        # request
        url = "/services/{}/resume".format(service_id)
        if force:
            url = '{}?force=true'
        success, response = self._client_api.gen_request(req_type="post",
                                                         path=url)
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def _get_bot_email(self, bot=None):

        if bot is None:
            project = self._project
            if project is None and self._project_id is not None:
                project = repositories.Projects(client_api=self._client_api).get(project_id=self._project_id)

            if project is None:
                raise exceptions.PlatformException(error='2001',
                                                   message='Need project entity or bot to perform this action')
            bots = project.bots.list()
            if len(bots) == 0:
                logger.info('Bot not found for project. Creating a default bot')
                bot = project.bots.create(name='default')
            else:
                bot = bots[0]
                if len(bots) > 1:
                    logger.debug('More than one bot users. Choosing first. email: {}'.format(bots[0].email))

        if isinstance(bot, str):
            bot_email = bot
        elif isinstance(bot, entities.Bot) or isinstance(bot, entities.User):
            bot_email = bot.email
        else:
            raise ValueError('input "bot" must be a str or a Bot type, got: {}'.format(type(bot)))

        return bot_email

    @staticmethod
    def _parse_init_input(init_input):
        if not isinstance(init_input, dict):
            if init_input is None:
                init_input = dict()
            else:
                init_params = dict()
                if not isinstance(init_input, list):
                    init_input = [init_input]
                for param in init_input:
                    if isinstance(param, entities.FunctionIO):
                        init_params.update(param.to_json(resource='service'))
                    else:
                        raise exceptions.PlatformException('400', 'Unknown type of init params')
                init_input = init_params

        return init_input

    def name_validation(self, name: str):
        """
        Validation service name.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str name: service name

        **Example**:

        .. code-block:: python

            package.services.name_validation(name='name')
        """
        url = '/piper-misc/naming/services/{}'.format(name)

        # request
        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url)
        if not success:
            raise exceptions.PlatformException(response)

    def _create(self,
                service_name: str = None,
                package: entities.Package = None,
                module_name: str = None,
                bot: Union[entities.Bot, str] = None,
                revision: str or int = None,
                init_input: Union[List[entities.FunctionIO], entities.FunctionIO, dict] = None,
                runtime: Union[entities.KubernetesRuntime, dict] = None,
                pod_type: entities.InstanceCatalog = None,
                project_id: str = None,
                sdk_version: str = None,
                agent_versions: dict = None,
                verify: bool = True,
                driver_id: str = None,
                run_execution_as_process: bool = None,
                execution_timeout: int = None,
                drain_time: int = None,
                on_reset: str = None,
                max_attempts: int = None,
                secrets=None,
                integrations=None,
                active: bool = True,
                **kwargs
                ) -> entities.Service:
        """
        Create service entity.


        :param str service_name: service name
        :param dtlpy.entities.package.Package package: package entity
        :param str module_name: module name
        :param str bot: bot email
        :param str revision: package revision - default=latest
        :param init_input: config to run at startup
        :param dict runtime: runtime resources
        :param str pod_type: pod type dl.InstanceCatalog
        :param str project_id: project id
        :param str sdk_version:  - optional - string - sdk version
        :param dict agent_versions: - dictionary - - optional -versions of sdk, agent runner and agent proxy
        :param bool verify: verify the inputs
        :param str driver_id: driver id
        :param bool run_execution_as_process: run execution as process
        :param int execution_timeout: execution timeout
        :param int drain_time: drain time
        :param str on_reset: on reset
        :param int max_attempts: Maximum execution retries in-case of a service reset
        :param bool force: optional - terminate old replicas immediately
        :param list secrets: list of the integrations ids
        :param list integrations: list of the integrations
        :param bool active: optional - if true, service will be active
        :param kwargs:
        :return: Service object
        :rtype: dtlpy.entities.service.Service
        """
        if package is None:
            if self._package is None:
                raise exceptions.PlatformException('400', 'Please provide param package')
            package = self._package

        if verify is not None:
            logger.warning('verify attribute has been deprecated and will be ignored')

        is_global = kwargs.get('is_global', None)
        jwt_forward = kwargs.get('jwt_forward', None)

        if is_global is not None or jwt_forward is not None:
            logger.warning(
                'Params jwt_forward and is_global are restricted to superuser. '
                'If you are not a superuser this action will not work'
            )

        service_config = dict()
        if package is not None and package.service_config is not None:
            service_config = package.service_config

        if agent_versions is None:
            if sdk_version is None:
                sdk_version = service_config.get('versions', dict()).get('dtlpy', __version__)
            agent_versions = {
                "dtlpy": sdk_version
            }

        if project_id is None:
            if self._project is None and self._project_id is None:
                raise exceptions.PlatformException('400', 'Please provide project id')
            elif self._project_id is not None:
                project_id = self._project_id
            elif self._project is not None:
                project_id = self._project.id

        if service_name is None:
            service_name = 'default-service'

        # payload
        payload = {
            'name': service_name,
            'projectId': project_id,
            'packageId': package.id,
            'initParams': self._parse_init_input(init_input=init_input),
            'botUserName': self._get_bot_email(bot=bot),
            'versions': agent_versions,
            'packageRevision': revision if revision is not None else package.version,
            'driverId': driver_id,
            'active': active
        }

        if secrets is not None:
            if not isinstance(secrets, list):
                secrets = [secrets]
            payload['secrets'] = secrets

        if integrations is not None:
            if not isinstance(integrations, list):
                integrations = [integrations]
            payload['integrations'] = integrations

        if runtime is not None:
            if isinstance(runtime, entities.KubernetesRuntime):
                runtime = runtime.to_json()

        if pod_type is not None:
            if runtime is None:
                runtime = {'podType': pod_type}
            else:
                runtime['podType'] = pod_type

        if runtime is not None:
            payload['runtime'] = runtime

        if module_name is not None:
            payload['moduleName'] = module_name

        if is_global is not None:
            payload['global'] = is_global

        if max_attempts is not None:
            payload['maxAttempts'] = max_attempts

        if jwt_forward is not None:
            payload['useUserJwt'] = jwt_forward

        if run_execution_as_process is not None:
            payload['runExecutionAsProcess'] = run_execution_as_process

        if drain_time is not None:
            payload['drainTime'] = drain_time

        if on_reset is not None:
            payload['onReset'] = on_reset

        if execution_timeout is not None:
            payload['executionTimeout'] = execution_timeout

        # request
        success, response = self._client_api.gen_request(
            req_type='post',
            path='/services',
            json_req=payload
        )

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Service.from_json(
            _json=response.json(),
            client_api=self._client_api,
            package=package,
            project=self._project
        )

    @_api_reference.add(path='/services/{id}', method='delete')
    def delete(self, service_name: str = None, service_id: str = None, force=False):
        """
        Delete Service object

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        You must provide at least ONE of the following params: service_id, service_name.

        :param force:
        :param str service_name: by name
        :param str service_id: by id
        :return: True
        :rtype: bool

        **Example**:

        .. code-block:: python

            is_deleted = package.services.delete(service_id='service_id')
        """
        # get bby name
        if service_id is None:
            if service_name is None:
                raise exceptions.PlatformException('400', 'Must provide either service id or service name')
            else:
                service_id = self.get(service_name=service_name).id

        path = "/services/{}".format(service_id)
        if force:
            path = '{}?force=true'.format(path)

        # request
        success, response = self._client_api.gen_request(
            req_type="delete",
            path=path
        )
        if not success:
            raise exceptions.PlatformException(response)
        return True

    @_api_reference.add(path='/services/{id}', method='patch')
    def update(self, service: entities.Service, force: bool = False) -> entities.Service:
        """
        Update service changes to platform.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param dtlpy.entities.service.Service service: Service entity
        :param bool force: optional - terminate old replicas immediately
        :return: Service entity
        :rtype: dtlpy.entities.service.Service

        **Example**:

        .. code-block:: python

            service = package.services.update(service='service_entity')
        """
        assert isinstance(service, entities.Service)

        # payload
        payload = service.to_json()

        # request
        url = '/services/{}'.format(service.id)
        if force:
            url = '{}?force=true'.format(url)
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url,
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        if self._package is not None:
            package = self._package
        else:
            package = service._package

        return entities.Service.from_json(_json=response.json(),
                                          client_api=self._client_api,
                                          package=package,
                                          project=self._project)

    def activate_slots(
            self,
            service: entities.Service,
            project_id: str = None,
            task_id: str = None,
            dataset_id: str = None,
            org_id: str = None,
            user_email: str = None,
            slots: List[entities.PackageSlot] = None,
            role=None,
            prevent_override: bool = True,
            visible: bool = True,
            icon: str = 'fas fa-magic',
            **kwargs
    ):
        """
        Activate service slots (creates buttons in the UI that activate services).

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param dtlpy.entities.service.Service service: service entity
        :param str project_id: project id
        :param str task_id: task id
        :param str dataset_id: dataset id
        :param str org_id: org id
        :param str user_email: user email
        :param list slots: list of entities.PackageSlot
        :param str role: user role MemberOrgRole.ADMIN, MemberOrgRole.owner, MemberOrgRole.MEMBER, MemberOrgRole.WORKER
        :param bool prevent_override: True to prevent override
        :param bool visible: visible
        :param str icon: icon
        :param kwargs: all additional arguments
        :return: list of user setting for activated slots
        :rtype: list

        **Example**:

        .. code-block:: python

            setting = package.services.activate_slots(service='service_entity',
                                            project_id='project_id',
                                            slots=List[entities.PackageSlot],
                                            icon='fas fa-magic')
        """
        package = service.package
        if not isinstance(package.slots, list) or len(package.slots) == 0:
            raise exceptions.PlatformException('400', "Service's package has no slots")

        if kwargs.get('is_global', False):
            project_id = '*'
            scope_ids = [project_id]
        else:
            scope_ids = [s_id for s_id in [project_id, task_id, org_id, dataset_id, user_email] if s_id is not None]
            if len(scope_ids) == 0:
                raise exceptions.PlatformException('400', "Must provide scope resource ID")

        settings = list()

        if role is None:
            role = entities.Role.ALL

        if not slots:
            slots = [s.to_json() for s in service.package.slots]
        elif isinstance(slots, list) and isinstance(slots[0], entities.PackageSlot):
            slots = [s.to_json() for s in slots]
        else:
            raise exceptions.PlatformException('400', "Slots param must be a list of PackageSlot objects")

        for scope_id in scope_ids:

            if kwargs.get('is_global', False):
                scope_type = entities.PlatformEntityType.DATALOOP
            elif scope_id == project_id:
                scope_type = entities.PlatformEntityType.PROJECT
            elif scope_id == task_id:
                scope_type = entities.PlatformEntityType.TASK
            elif scope_id == dataset_id:
                scope_type = entities.PlatformEntityType.DATASET
            elif scope_id == user_email:
                scope_type = entities.PlatformEntityType.USER
            elif scope_id == org_id:
                scope_type = entities.PlatformEntityType.ORG
            else:
                raise exceptions.PlatformException('400', "Unknown resource id")

            setting = entities.Setting(
                default_value=True,
                value=True,
                inputs=None,
                name=service.name,
                value_type=entities.SettingsValueTypes.BOOLEAN,
                scope=entities.SettingScope(
                    type=scope_type,
                    id=scope_id,
                    role=role,
                    prevent_override=prevent_override,
                    visible=visible
                ),
                metadata={
                    'serviceId': service.id,
                    'serviceName': service.name,
                    'projectId': service.project_id,
                    'slots': slots
                },
                description=service.name,
                icon=icon,
                section_name=entities.SettingsSectionNames.APPLICATIONS,
                sub_section_name=None,
                hint=None
            )

            try:
                settings.append(self._settings.create(setting=setting))
            except exceptions.BadRequest as err:
                if 'settings already exist' in err.message:
                    old_setting = self._settings.get(setting_name=setting.name)
                    setting.id = old_setting.id
                    settings.append(self._settings.update(setting=setting))
                else:
                    raise err

        return settings

    @_api_reference.add(path='/services/logs', method='post')
    def log(self,
            service,
            size=100,
            checkpoint=None,
            start=None,
            end=None,
            follow=False,
            text=None,
            execution_id=None,
            function_name=None,
            replica_id=None,
            system=False,
            view=True,
            until_completed=True,
            log_level='DEBUG',
            model_id: str = None,
            model_operation: str = None,
            project_id: str = None
            ):
        """
        Get service logs.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param dtlpy.entities.service.Service service: service object
        :param int size: size
        :param dict checkpoint: the information from the lst point checked in the service
        :param str start: iso format time
        :param str end: iso format time
        :param bool follow: if true, keep stream future logs
        :param str text: text
        :param str execution_id: execution id
        :param str function_name: function name
        :param str replica_id: replica id
        :param bool system: system
        :param bool view: if true, print out all the logs
        :param bool until_completed: wait until completed
        :param str log_level: the log level to display dl.LoggingLevel
        :param str model_id: model id
        :param str model_operation: model operation action
        :param str project_id: project id
        :return: ServiceLog entity
        :rtype: ServiceLog

        **Example**:

        .. code-block:: python

            service_logs = package.services.log(service='service_entity')
        """
        if not isinstance(service, entities.Service) and model_id is None:
            raise exceptions.PlatformException('400', 'Must provide service or model_id')
        if isinstance(log_level, str):
            log_level = log_level.upper()

        payload = {
            'direction': 'asc',
            'follow': follow,
            'system': system,
            'logLevel': log_level
        }

        if service is not None:
            payload['serviceId'] = service.id

        if size is not None:
            payload['size'] = size

        if execution_id is not None:
            payload['executionId'] = [execution_id]

        if function_name is not None:
            payload['functionName'] = function_name

        if text is not None:
            payload['text'] = text

        if replica_id is not None:
            payload['replicaId'] = replica_id

        if checkpoint is not None:
            payload['checkpoint'] = checkpoint

        if start is not None:
            payload['start'] = start
        else:
            payload['start'] = datetime.datetime(datetime.date.today().year,
                                                 datetime.date.today().month,
                                                 datetime.date.today().day,
                                                 0,
                                                 0,
                                                 0).isoformat()

        if end is not None:
            payload['end'] = end

        if model_id is not None:
            payload['modelId'] = model_id

        if model_operation is not None:
            payload['modelOperation'] = model_operation

        if project_id is not None:
            payload['projectId'] = project_id
        else:
            payload['projectId'] = service.project_id

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/services/logs',
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        log = ServiceLog(_json=response.json(),
                         service=service,
                         services=self,
                         start=payload['start'],
                         follow=follow,
                         execution_id=execution_id,
                         function_name=function_name,
                         replica_id=replica_id,
                         system=system,
                         model_id=model_id,
                         model_operation=model_operation,
                         project_id=project_id)

        if view:
            log.view(until_completed=until_completed)
        else:
            return log

    def execute(self,
                service: entities.Service = None,
                service_id: str = None,
                service_name: str = None,
                sync: bool = False,
                function_name: str = None,
                stream_logs: bool = False,
                execution_input=None,
                resource=None,
                item_id=None,
                dataset_id=None,
                annotation_id=None,
                project_id=None,
                ) -> entities.Execution:
        """
        Execute a function on an existing service.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param dtlpy.entities.service.Service service:  service entity
        :param str service_id:  service id
        :param str service_name: service name
        :param bool sync: wait for function to end
        :param str function_name: function name to run
        :param bool stream_logs: prints logs of the new execution. only works with sync=True
        :param execution_input: input dictionary or list of FunctionIO entities
        :param str resource: dl.PackageInputType - input type.
        :param str item_id: str - optional - input to function
        :param str dataset_id: str - optional - input to function
        :param str annotation_id: str - optional - input to function
        :param str project_id: str - resource's project
        :return: entities.Execution
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            execution = package.services.execute(service='service_entity',
                                    function_name='run',
                                    item_id='item_id',
                                    project_id='project_id')
        """
        if service is None:
            service = self.get(service_id=service_id, service_name=service_name)
        execution = repositories.Executions(service=service,
                                            client_api=self._client_api,
                                            project=self._project).create(service_id=service.id,
                                                                          sync=sync,
                                                                          execution_input=execution_input,
                                                                          function_name=function_name,
                                                                          resource=resource,
                                                                          item_id=item_id,
                                                                          dataset_id=dataset_id,
                                                                          annotation_id=annotation_id,
                                                                          project_id=project_id,
                                                                          stream_logs=stream_logs)
        return execution

    @_api_reference.add(path='/services', method='post')
    def deploy(self,
               service_name: str = None,
               package: entities.Package = None,
               bot: Union[entities.Bot, str] = None,
               revision: str or int = None,
               init_input: Union[List[entities.FunctionIO], entities.FunctionIO, dict] = None,
               runtime: Union[entities.KubernetesRuntime, dict] = None,
               pod_type: entities.InstanceCatalog = None,
               sdk_version: str = None,
               agent_versions: dict = None,
               verify: bool = True,
               checkout: bool = False,
               module_name: str = None,
               project_id: str = None,
               driver_id: str = None,
               func: Callable = None,
               run_execution_as_process: bool = None,
               execution_timeout: int = None,
               drain_time: int = None,
               max_attempts: int = None,
               on_reset: str = None,
               force: bool = False,
               secrets: list = None,
               integrations: list = None,
               active: bool = True,
               **kwargs) -> entities.Service:
        """
        Deploy service.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param str service_name: name
        :param dtlpy.entities.package.Package package: package entity
        :param str bot: bot email
        :param str revision: package revision of version
        :param init_input: config to run at startup
        :param dict runtime: runtime resources
        :param str pod_type: pod type dl.InstanceCatalog
        :param str sdk_version:  - optional - string - sdk version
        :param str agent_versions: - dictionary - - optional -versions of sdk
        :param bool verify: if true, verify the inputs
        :param bool checkout: if true, checkout (switch) to service
        :param str module_name: module name
        :param str project_id: project id
        :param str driver_id: driver id
        :param Callable func: function to deploy
        :param bool run_execution_as_process: if true, run execution as process
        :param int execution_timeout: execution timeout in seconds
        :param int drain_time: drain time in seconds
        :param int max_attempts: maximum execution retries in-case of a service reset
        :param str on_reset: what happens on reset
        :param bool force: optional - if true, terminate old replicas immediately
        :param list secrets: list of the integrations ids
        :param list integrations: list of the integrations
        :param bool active: if true, activate the service
        :param kwargs: list of additional arguments
        :return: Service object
        :rtype: dtlpy.entities.service.Service

        **Example**:

        .. code-block:: python

            service = package.services.deploy(service_name=package_name,
                                    execution_timeout=3 * 60 * 60,
                                    module_name=module.name,
                                    runtime=dl.KubernetesRuntime(
                                        concurrency=10,
                                        pod_type=dl.InstanceCatalog.REGULAR_S,
                                        autoscaler=dl.KubernetesRabbitmqAutoscaler(
                                            min_replicas=1,
                                            max_replicas=20,
                                            queue_length=20
                                        )
                                    )
                                )
        """
        if package is None and isinstance(package, entities.Dpk):
            raise exceptions.PlatformException('400', 'cannot deploy dpk package. Please install the app')
        package = package if package is not None else self._package
        if service_name is None:
            get_name = False
            if package is not None and package.service_config is not None and 'name' in package.service_config:
                service_name = package.service_config['name']
                get_name = True
            else:
                if package is not None:
                    service_name = package.name
                else:
                    service_name = 'default-service'
            if not get_name:
                logger.warning('service_name not provided, using: {} by default'.format(service_name))

        if isinstance(revision, int):
            logger.warning('Deprecation Warning - Package/service versions have been refactored'
                           'The version you provided has type: int, it will be converted to: 1.0.{}'
                           'Next time use a 3-level semver for package/service versions'.format(revision))

        if func is not None:
            package = self.__deploy_function(name=service_name, project=self._project, func=func)

        if init_input is not None and not isinstance(init_input, dict):
            if not isinstance(init_input, list):
                init_input = [init_input]

            if len(init_input) > 0 and isinstance(init_input[0], entities.FunctionIO):
                params = dict()
                for i_param, param in enumerate(init_input):
                    params[param.name] = param.value
                init_input = params
            elif len(init_input) == 0:
                init_input = None
            else:
                raise exceptions.PlatformException(
                    error='400',
                    message='Unknown init_input type. expecting list or dict, got: {}'.format(type(init_input))
                )

        if project_id is None:
            if self._project is not None:
                project_id = self._project.id
            else:
                project_id = self._project_id

        filters = entities.Filters(resource=entities.FiltersResource.SERVICE, use_defaults=False)
        filters.add(field='name', values=service_name)
        if project_id is not None:
            filters.add(field='projectId', values=project_id)
        services = self.list(filters=filters)
        if services.items_count > 1:
            raise exceptions.PlatformException('400',
                                               'More than 1 service by this name are associated with this user. '
                                               'Please provide project_id')
        elif services.items_count > 0:
            service = services.items[0]
            if runtime is not None:
                service.runtime = runtime
            if init_input is not None:
                service.init_input = init_input
            if revision is not None:
                service.package_revision = revision
            if agent_versions is not None:
                service.versions = agent_versions
            elif sdk_version:
                service.versions = {'dtlpy': sdk_version}
            if driver_id is not None:
                service.driver_id = driver_id
            if secrets is not None:
                if not isinstance(secrets, list):
                    secrets = [secrets]
                service.secrets = secrets
            if integrations is not None:
                if not isinstance(integrations, list):
                    integrations = [integrations]
                service.integrations = integrations
            service = self.update(service=service, force=force)
        else:
            service = self._create(service_name=service_name,
                                   package=package,
                                   project_id=project_id,
                                   bot=bot,
                                   revision=revision,
                                   init_input=init_input,
                                   runtime=runtime,
                                   pod_type=pod_type,
                                   sdk_version=sdk_version,
                                   agent_versions=agent_versions,
                                   verify=verify,
                                   module_name=module_name,
                                   driver_id=driver_id,
                                   jwt_forward=kwargs.get('jwt_forward', None),
                                   is_global=kwargs.get('is_global', None),
                                   run_execution_as_process=run_execution_as_process,
                                   execution_timeout=execution_timeout,
                                   drain_time=drain_time,
                                   max_attempts=max_attempts,
                                   on_reset=on_reset,
                                   secrets=secrets,
                                   integrations=integrations,
                                   active=active
                                   )
        if checkout:
            self.checkout(service=service)
        return service

    @staticmethod
    def __get_import_string(imports: List[str]):
        import_string = ''
        if imports is not None:
            import_string = '\n'.join(imports)
        return import_string

    @staticmethod
    def __get_inputs(func):
        method = inspect.signature(func)
        params = list(method.parameters)
        inpts = list()
        inputs_types = {i.name.lower(): i.value for i in list(entities.PackageInputType)}
        for arg in params:
            if arg in inputs_types:
                inpt_type = inputs_types[arg]
            else:
                inpt_type = entities.PackageInputType.JSON
            inpts.append(entities.FunctionIO(type=inpt_type, name=arg))
        return inpts

    def __deploy_function(self,
                          name: str,
                          func: Callable,
                          project: entities.Project) -> entities.Package:
        package_dir = tempfile.mkdtemp()
        # imports_string = self.__get_import_string()
        imports_string = ''

        main_file = os.path.join(package_dir, entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT)
        with open(assets.paths.PARTIAL_MAIN_FILEPATH, 'r') as f:
            main_string = f.read()
        lines = inspect.getsourcelines(func)
        tabs_diff = lines[0][0].count('    ') - 1
        for line_index in range(len(lines[0])):
            line_tabs = lines[0][line_index].count('    ') - tabs_diff
            lines[0][line_index] = ('    ' * line_tabs) + lines[0][line_index].strip() + '\n'

        method_func_string = "".join(lines[0])

        with open(main_file, 'w') as f:
            f.write('{}\n{}\n    @staticmethod\n{}'.format(imports_string, main_string,
                                                           method_func_string))

        function = entities.PackageFunction(name=func.__name__, inputs=self.__get_inputs(func=func))
        module = entities.PackageModule(functions=[function],
                                        entry_point=entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT)
        packages = repositories.Packages(client_api=self._client_api, project=project)
        return packages.push(src_path=package_dir,
                             package_name=name,
                             checkout=True,
                             modules=[module])

    def deploy_from_local_folder(self,
                                 cwd=None,
                                 service_file=None,
                                 bot=None,
                                 checkout=False,
                                 force=False
                                 ) -> entities.Service:
        """
        Deploy from local folder in local environment.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param str cwd: optional - package working directory. Default=cwd
        :param str service_file: optional - service file. Default=None
        :param str bot: bot
        :param checkout: checkout
        :param bool force: optional - terminate old replicas immediately
        :return: Service object
        :rtype: dtlpy.entities.service.Service

        **Example**:

        .. code-block:: python

            service = package.services.deploy_from_local_folder(cwd='file_path',
                                                      service_file='service_file')
        """
        # get cwd and service.json path
        if cwd is None:
            cwd = os.getcwd()
        if service_file is None:
            service_file = os.path.join(cwd, assets.paths.SERVICE_FILENAME)

        # load service json
        if os.path.isfile(service_file):
            with open(service_file, 'r') as f:
                service_json = json.load(f)
            service_triggers = service_json.get('triggers', list())
        else:
            raise exceptions.PlatformException(error='400',
                                               message='Could not find service.json in path: {}'.format(cwd))

        # get package
        package_name = service_json.get('packageName', None)
        packages = repositories.Packages(client_api=self._client_api, project=self._project)

        if package_name is None:
            package = packages.get()
        else:
            package = packages.get(package_name=package_name)

        name = service_json.get('name', None)
        revision = service_json.get('revision', package.version)
        init_input = service_json.get('initParams', dict())
        runtime = service_json.get('runtime', dict())
        sdk_version = service_json.get('version', None)
        agent_versions = service_json.get('versions', None)
        verify = service_json.get('verify', True)
        module_name = service_json.get('module_name', None)
        run_execution_as_process = service_json.get('run_execution_as_process', None)
        execution_timeout = service_json.get('execution_timeout', None)
        drain_time = service_json.get('drain_time', None)
        on_reset = service_json.get('on_reset', None)
        max_attempts = service_json.get('maxAttempts', None)

        service = self.deploy(bot=bot,
                              service_name=name,
                              package=package,
                              revision=revision,
                              runtime=runtime,
                              init_input=init_input,
                              sdk_version=sdk_version,
                              agent_versions=agent_versions,
                              verify=verify,
                              checkout=checkout,
                              run_execution_as_process=run_execution_as_process,
                              execution_timeout=execution_timeout,
                              drain_time=drain_time,
                              max_attempts=max_attempts,
                              on_reset=on_reset,
                              module_name=module_name,
                              force=force
                              )

        logger.info('Service was deployed successfully. Service id: {}'.format(service.id))

        if len(service_triggers) > 0:
            logger.info('Creating triggers...')
            triggers = repositories.Triggers(client_api=self._client_api, project=self._project)

            for trigger in service_triggers:
                name = trigger.get('name', None)
                filters = trigger.get('filter', dict())
                resource = trigger['resource']
                actions = trigger.get('actions', list())
                active = trigger.get('active', True)
                execution_mode = trigger.get('executionMode', None)
                function_name = trigger.get('function', None)

                trigger = triggers.create(service_id=service.id,
                                          name=name,
                                          filters=filters,
                                          resource=resource,
                                          actions=actions,
                                          active=active,
                                          execution_mode=execution_mode,
                                          function_name=function_name)

                logger.info('Trigger was created successfully. Service id: {}'.format(trigger.id))

        logging.info('Successfully deployed!')
        return service

    def __enable_cache(self,
                       url,
                       organization: entities.Organization,
                       pod_type=entities.PodType.SMALL):
        payload = {
            "org": {
                "name": organization.name,
                "id": organization.id
            },
            "runner": {
                "podType": pod_type  # small, medium, high
            }
        }

        return self._client_api.gen_request(req_type='post',
                                            path=url,
                                            json_req=payload)

    def __polling_wait(self, organization, pod_type, backoff_factor=1):
        fs_url_path = '/services/fs-cache?mode={}'.format('get')
        i = 1
        while True:
            success, response = self.__enable_cache(url=fs_url_path, organization=organization, pod_type=pod_type)
            if response.json().get('state', None) == 'READY':
                break
            sleep_time = min(backoff_factor * (2 ** (i - 1)), MAX_WAIT_TIME)
            logger.debug("Going to sleep {:.2f}[s]".format(sleep_time))
            time.sleep(sleep_time)
            i += 1
        return success

    def _cache_action(self,
                      organization: entities.Organization = None,
                      mode=entities.CacheAction.APPLY,
                      pod_type=entities.PodType.SMALL):
        """
        Add or remove Cache for the org

        **Prerequisites**: You must be an organization *owner*

        You must provide at least ONE of the following params: organization, organization_name, or organization_id.

        :param entities.Organization organization: Organization object
        :param str mode: dl.CacheAction.APPLY or dl.CacheAction.DESTROY
        :param entities.PodType pod_type:  dl.PodType.SMALL, dl.PodType.MEDIUM, dl.PodType.HIGH
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = dl.organizations.enable_cache(organization='organization',
                                          mode=dl.CacheAction.APPLY)
        """
        if organization is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        fs_mode = mode if mode != entities.CacheAction.APPLY else '{}-filestore'.format(mode)
        apply_fs_url_path = '/services/fs-cache?mode={}'.format(fs_mode)
        apply_volume_url_path = '/services/fs-cache?mode={}'.format(mode)
        cache_url_path = '/services/cache?mode={}'.format(mode)

        success, response = self.__enable_cache(url=apply_fs_url_path, organization=organization, pod_type=pod_type)
        if not success:
            raise exceptions.PlatformException(response)

        if mode == entities.CacheAction.APPLY:
            self.__polling_wait(organization=organization, pod_type=pod_type)
            success, response = self.__enable_cache(url=apply_volume_url_path, organization=organization,
                                                    pod_type=pod_type)
            if not success:
                raise exceptions.PlatformException(response)

        success, response = self.__enable_cache(url=cache_url_path, organization=organization, pod_type=pod_type)
        if not success:
            raise exceptions.PlatformException(response)

        return True

    def restart(self, service: entities.Service, replica_name: str = None):
        """
        Restart service replica

        **Prerequisites**: You must be in the role of a *developer*.

        :param dtlpy.entities.service.Service service: service entity
        :param str replica_name: replica name
        :return: True
        :rtype: bool

        **Example**:

        .. code-block:: python

            is_restarted = dl.services.restart(service='service_entity',
                                                    replica_name='replica_name')
        """
        payload = {}

        if replica_name is not None:
            payload['replicaName'] = replica_name

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/services/{}/restart'.format(service.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        return True


class ServiceLog:
    """
    Service Log
    """

    def __init__(self,
                 _json: dict,
                 service: entities.Service,
                 services: Services,
                 start=None,
                 follow=None,
                 execution_id=None,
                 function_name=None,
                 replica_id=None,
                 system=False,
                 model_id=None,
                 model_operation=None,
                 project_id=None):

        self.logs = _json.get('logs', dict())
        self.checkpoint = _json.get('checkpoint', None)
        self.stop = _json.get('stop', False)
        self.service = service
        self.services = services
        self.start = start
        self.follow = follow
        self.execution_id = execution_id
        self.function_name = function_name
        self.replica_id = replica_id
        self.system = system
        self.model_id = model_id
        self.model_operation = model_operation
        self.project_id = project_id

    def get_next_log(self):
        log = self.services.log(service=self.service,
                                checkpoint=self.checkpoint,
                                start=self.start,
                                follow=self.follow,
                                execution_id=self.execution_id,
                                function_name=self.function_name,
                                replica_id=self.replica_id,
                                system=self.system,
                                view=False,
                                model_id=self.model_id,
                                model_operation=self.model_operation,
                                project_id=self.project_id)

        self.logs = log.logs
        self.checkpoint = log.checkpoint
        self.stop = log.stop

    def view(self, until_completed):
        """
        View logs

        :param until_completed:
        """
        try:
            for log in self:
                print(log)
                if until_completed and FUNCTION_END_LINE in log:
                    break
        except KeyboardInterrupt:
            return

    def __iter__(self):
        while not self.stop:
            for log in self.logs:
                yield '{}: {}'.format(log.get('timestamp', self.start), log.get('message', '').strip())
            self.get_next_log()




from collections import namedtuple
from typing import List, Union
import traceback
import enum

from .. import entities, repositories, exceptions
from ..services.api_client import ApiClient


class SlotType(str, enum.Enum):
    ITEM_VIEWER = 'itemViewer'
    FLOATING_WINDOW = 'floatingWindow'


DEFAULT_STOPS = {SlotType.ITEM_VIEWER: {"type": "itemViewer",
                                        "configuration": {"layout": {"leftBar": True,
                                                                     "rightBar": True,
                                                                     "bottomBar": True,
                                                                     }
                                                          }
                                        },
                 SlotType.FLOATING_WINDOW: {"type": "floatingWindow",
                                            "configuration": {"layout": {"width": 455,
                                                                         "height": 340,
                                                                         "resizable": True,
                                                                         "backgroundColor": "dl-color-studio-panel"
                                                                         }
                                                              }
                                            }
                 }


class Slot(entities.DlEntity):
    type: str = entities.DlProperty(location=['type'], _type=str)
    configuration: dict = entities.DlProperty(location=['configuration'], _type=dict)

    def to_json(self) -> dict:
        return self._dict.copy()

    @classmethod
    def from_json(cls, _json):
        return cls(_dict=_json)


class Toolbar(entities.DlEntity):
    name: str = entities.DlProperty(location=['name'], _type=str)
    display_name: str = entities.DlProperty(location=['displayName'], _type=str)
    conditions: dict = entities.DlProperty(location=['conditions'], _type=dict)
    invoke: dict = entities.DlProperty(location=['invoke'], _type=dict)
    location: str = entities.DlProperty(location=['location'], _type=str)
    icon: str = entities.DlProperty(location=['icon'], _type=str)
    action: str = entities.DlProperty(location=['action'], _type=str, default=None)

    def to_json(self) -> dict:
        return self._dict.copy()

    @classmethod
    def from_json(cls, _json):
        return cls(_dict=_json)


class Panel(entities.DlEntity):
    name: str = entities.DlProperty(location=['name'], _type=str)
    path: str = entities.DlProperty(location=['path'], _type=str, default=None)
    min_role: list = entities.DlProperty(location=['minRole'], _type=list)
    supported_slots: list = entities.DlProperty(location=['supportedSlots'], _type=list)

    metadata = entities.DlProperty(location=['metadata'], _type=list)
    default_settings = entities.DlProperty(location=['defaultSettings'], _type=list)
    conditions = entities.DlProperty(location=['conditions'], _type=list)

    def to_json(self) -> dict:
        return self._dict.copy()

    @classmethod
    def from_json(cls, _json):
        return cls(_dict=_json)


class ToolbarInvoke(entities.DlEntity):
    type: str = entities.DlProperty(location=['type'], _type=str)
    panel: str = entities.DlProperty(location=['panel'], _type=str)
    namespace: str = entities.DlProperty(location=['namespace'], _type=str)
    source: str = entities.DlProperty(location=['source'], _type=str)
    input_options: dict = entities.DlProperty(location=['inputOptions'], _type=dict)

    def to_json(self) -> dict:
        return self._dict.copy()

    @classmethod
    def from_json(cls, _json):
        return cls(_dict=_json)


class CustomNodeScope(str, enum.Enum):
    GLOBAL = "global",
    PROJECT = "project",
    NODE = 'node'


class PipelineNode(entities.DlEntity):
    display_name: str = entities.DlProperty(location=['displayName'], _type=str)
    panel: str = entities.DlProperty(location=['panel'], _type=str)
    invoke: ToolbarInvoke = entities.DlProperty(location=['invoke'], _kls='ToolbarInvoke')
    icon: str = entities.DlProperty(location=['icon'], _type=str)
    categories: List[str] = entities.DlProperty(location=['categories'], _type=list)
    description: str = entities.DlProperty(location=['description'], _type=str)
    configuration: dict = entities.DlProperty(location=['configuration'], _type=dict)
    scope: CustomNodeScope = entities.DlProperty(location=['scope'], _type=CustomNodeScope)
    compute_config: str = entities.DlProperty(location=['computeConfig'], _type=str, default=None)

    def to_json(self) -> dict:
        return self._dict.copy()

    @classmethod
    def from_json(cls, _json):
        return cls(_dict=_json)


class DpkComputeConfig(entities.DlEntity):
    run_execution_as_process: bool = entities.DlProperty(location=['runExecutionAsProcess'], _type=bool)
    execution_timeout: int = entities.DlProperty(location=['executionTimeout'], _type=int)
    drain_time: int = entities.DlProperty(location=['drainTime'], _type=int)
    on_reset: str = entities.DlProperty(location=['onReset'], _type=str)
    runtime: dict = entities.DlProperty(location=['runtime'], _type=dict)
    bot_user_name: str = entities.DlProperty(location=['botUserName'], _type=str)
    max_attempts: int = entities.DlProperty(location=['maxAttempts'], _type=int)
    use_user_jwt: bool = entities.DlProperty(location=['useUserJwt'], _type=bool)
    driver_id: str = entities.DlProperty(location=['driverId'], _type=str)
    versions: dict = entities.DlProperty(location=['versions'], _type=dict)
    name: str = entities.DlProperty(location=['name'], _type=str)
    integrations: List[dict] = entities.DlProperty(location=['integrations'], _type=list)

    def to_json(self) -> dict:
        return self._dict.copy()

    @classmethod
    def from_json(cls, _json):
        return cls(_dict=_json)


class DpkComponentChannel(entities.DlEntity):
    name: str = entities.DlProperty(location=['name'], _type=str)
    icon: str = entities.DlProperty(location=['icon'], _type=str)
    description: str = entities.DlProperty(location=['description'], _type=str)
    is_global: bool = entities.DlProperty(location=['global'], _type=bool)
    metadata: dict = entities.DlProperty(location=['metadata'], _type=dict)
    context: dict = entities.DlProperty(location=['context'], _type=dict)
    filters: List[dict] = entities.DlProperty(location=['filters'], _type=list)

    def to_json(self) -> dict:
        return self._dict.copy()

    @classmethod
    def from_json(cls, _json):
        return cls(_dict=_json)


class Components(entities.DlEntity):
    panels: List[Panel] = entities.DlProperty(location=['panels'], _kls='Panel')
    modules: List[entities.PackageModule] = entities.DlProperty(location=['modules'], _kls='PackageModule')
    services: List[dict] = entities.DlProperty(location=['services'])
    triggers: List[dict] = entities.DlProperty(location=['triggers'])
    pipeline_nodes: List[PipelineNode] = entities.DlProperty(location=['pipelineNodes'])
    toolbars: List[Toolbar] = entities.DlProperty(location=['toolbars'], _kls='Toolbar')
    models: List[dict] = entities.DlProperty(location=['models'])
    compute_configs: List[DpkComputeConfig] = entities.DlProperty(location=['computeConfigs'], _kls='DpkComputeConfig')
    channels: List[DpkComponentChannel] = entities.DlProperty(location=['channels'], _kls='DpkComponentChannel')
    pipeline_templates: List[dict] = entities.DlProperty(location=['pipelineTemplates'])
    integrations: List[dict] = entities.DlProperty(location=['integrations'])

    @panels.default
    def default_panels(self):
        self._dict['panels'] = list()
        return self._dict['panels']

    @modules.default
    def default_modules(self):
        self._dict['modules'] = list()
        return self._dict['modules']

    @services.default
    def default_services(self):
        self._dict['services'] = list()
        return self._dict['services']

    @triggers.default
    def default_triggers(self):
        self._dict['triggers'] = list()
        return self._dict['triggers']

    @pipeline_nodes.default
    def default_pipelines(self):
        self._dict['pipelines'] = list()
        return self._dict['pipelines']

    @toolbars.default
    def default_toolbars(self):
        self._dict['toolbars'] = list()
        return self._dict['toolbars']

    @models.default
    def default_models(self):
        self._dict['models'] = list()
        return self._dict['models']

    @compute_configs.default
    def default_compute_configs(self):
        self._dict['computeConfigs'] = list()
        return self._dict['computeConfigs']

    @channels.default
    def default_channels(self):
        self._dict['channels'] = list()
        return self._dict['channels']

    @pipeline_templates.default
    def default_pipeline_templates(self):
        self._dict['pipelineTemplates'] = list()
        return self._dict['pipelineTemplates']

    @classmethod
    def from_json(cls, _json):
        inst = cls(_dict=_json)
        return inst

    def to_json(self):
        return self._dict.copy()


class Dpk(entities.DlEntity):
    # name change
    id: str = entities.DlProperty(location=['id'], _type=str)
    base_id: str = entities.DlProperty(location=['baseId'], _type=str)
    name: str = entities.DlProperty(location=['name'], _type=str)
    version: str = entities.DlProperty(location=['version'], _type=str)
    attributes: dict = entities.DlProperty(location=['attributes'], _type=dict)
    created_at: str = entities.DlProperty(location=['createdAt'], _type=str)
    updated_at: str = entities.DlProperty(location=['updatedAt'], _type=str)
    creator: str = entities.DlProperty(location=['creator'], _type=str)
    display_name: str = entities.DlProperty(location=['displayName'], _type=str)
    icon: str = entities.DlProperty(location=['icon'], _type=str)
    tags: list = entities.DlProperty(location=['tags'], _type=list)
    codebase: Union[entities.Codebase, None] = entities.DlProperty(location=['codebase'], _kls="Codebase")
    scope: str = entities.DlProperty(location=['scope'], _type=str)
    context: dict = entities.DlProperty(location=['context'], _type=dict)
    metadata: dict = entities.DlProperty(location=['metadata'], _type=dict)
    dependencies: dict = entities.DlProperty(location=['dependencies'], _type=List[dict])

    # defaults
    components: Components = entities.DlProperty(location=['components'], _kls='Components')
    description: str = entities.DlProperty(location=['description'], _type=str)
    url: str = entities.DlProperty(location=['url'], _type=str)

    # sdk
    client_api: ApiClient
    project: entities.Project
    _revisions = None
    __repositories = None

    @components.default
    def default_components(self):
        self._dict['components'] = dict()
        return self._dict['components']

    ################
    # repositories #
    ################
    @property
    def _repositories(self):
        if self.__repositories is None:
            reps = namedtuple('repositories',
                              field_names=['dpks', 'codebases', 'organizations', 'services'])

            self.__repositories = reps(
                dpks=repositories.Dpks(client_api=self.client_api, project=self.project),
                codebases=repositories.Codebases(client_api=self.client_api),
                organizations=repositories.Organizations(client_api=self.client_api),
                services=repositories.Services(client_api=self.client_api, project=self.project, package=self),
            )

        return self.__repositories

    @property
    def codebases(self):
        assert isinstance(self._repositories.codebases, repositories.Codebases)
        return self._repositories.codebases

    @property
    def organizations(self):
        assert isinstance(self._repositories.organizations, repositories.Organizations)
        return self._repositories.organizationsFapp

    @property
    def dpks(self) -> 'repositories.Dpks':
        assert isinstance(self._repositories.dpks, repositories.Dpks)
        return self._repositories.dpks

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    ###########
    # methods #
    ###########
    def publish(self):
        """
        Publish the dpk to Dataloop platform.

        **Example**
        .. code-block:: python
            published_dpk = dpk.publish()
        """
        return self.dpks.publish(dpk=self)

    def update(self):
        """
        Update the dpk attributes to Dataloop platform.

        **Example**
        .. code-block:: python
            updated_dpk = dpk.update()
        """
        return self.dpks.update(dpk=self)

    def pull(self, local_path):
        """
        Pulls the app from the platform as dpk file.

        Note: you must pass either dpk_name or dpk_id to the function.
        :param local_path: the path where you want to install the dpk file.
        :return local path where the package pull

        **Example**
        ..code-block:: python
            path = dl.dpks.pull(dpk_name='my-app')
        """
        return self.dpks.pull(dpk=self, local_path=local_path)

    def delete(self):
        """
        Delete the dpk from the app store.

        Note: after removing the dpk, you cant get it again, it's advised to pull it first.

        :return whether the operation ran successfully
        :rtype bool
        """
        return self.dpks.delete(self.id)

    def _get_revision_pages(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        returns the available versions of the dpk.

        :param entities.Filters filters: the filters to apply to the search.
        :return the available versions of the dpk.

        ** Example **
        ..code-block:: python
            versions = dl.dpks.revisions(dpk_name='name')
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.DPK)
        elif not isinstance(filters, entities.Filters):
            raise ValueError('Unknown filters type: {!r}'.format(type(filters)))
        elif filters.resource != entities.FiltersResource.DPK:
            raise TypeError('Filters resource must to be FiltersResource.DPK. Got: {!r}'.format(filters.resource))

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self.client_api,
                                       list_function=self._list_revisions)
        paged.get_page()
        return paged

    def _build_entities_from_response(self, response_items):
        return self.dpks._build_entities_from_response(response_items=response_items)

    def _list_revisions(self, filters: entities.Filters):
        url = '/app-registry/{}/revisions'.format(self.name)
        # request
        success, response = self.client_api.gen_request(req_type='post',
                                                        path=url,
                                                        json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @property
    def revisions(self):
        """
        Returns the available versions of the dpk.

        :return List[Dpk]

        ** Example **
        ..code-block:: python
        versions = dpk.revisions
        """
        if self._revisions is None:
            self._revisions = self._get_revision_pages()
        return self._revisions

    def get_revisions(self, version: str):
        """
        Get the dpk with the specified version.

        :param str version: the version of the dpk to get.
        :return: Dpk

        ** Example **
        ..code-block:: python
        dpk = dpk.get_revisions(version='1.0.0')
        """
        return self.dpks.get_revisions(dpk_id=self.base_id, version=version)

    @staticmethod
    def _protected_from_json(_json, client_api, project, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param _json:  platform json
        :param client_api: ApiClient entity
        :return:
        """
        try:
            package = Dpk.from_json(_json=_json,
                                    client_api=client_api,
                                    project=project,
                                    is_fetched=is_fetched)
            status = True
        except Exception:
            package = traceback.format_exc()
            status = False
        return status, package

    def to_json(self):
        """
        convert the class to json
        """
        _json = self._dict.copy()
        return _json

    @classmethod
    def from_json(cls,
                  _json,
                  client_api: ApiClient = None,
                  project: entities.Project = None,
                  is_fetched=True) -> 'Dpk':
        """
        Turn platform representation of app into an app entity

        :param dict _json: platform representation of package
        :param dl.ApiClient client_api: ApiClient entity
        :param dl.entities.Project project: The project where the dpk resides
        :param bool is_fetched: is Entity fetched from Platform
        :return: App entity
        :rtype: dtlpy.entities.App
        """
        res = cls(
            _dict=_json,
            client_api=client_api,
            project=project,
            is_fetched=is_fetched
        )

        return res



import mimetypes
import traceback
import datetime
import webvtt
import logging
import attr
import json
import os

import numpy as np
from PIL import Image

from .. import entities, PlatformException, miscellaneous

logger = logging.getLogger(name='dtlpy')


@attr.s
class AnnotationCollection(entities.BaseEntity):
    """
        Collection of Annotation entity
    """
    item = attr.ib(default=None)
    annotations = attr.ib()  # type: miscellaneous.List[entities.Annotation]
    _dataset = attr.ib(repr=False, default=None)
    _colors = attr.ib(repr=False, default=None)

    @property
    def dataset(self):
        if self._dataset is None:
            self._dataset = self.item.dataset
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @annotations.default
    def set_annotations(self):
        return list()

    def __iter__(self):
        for annotation in self.annotations:
            yield annotation

    def __getitem__(self, index: int):
        return self.annotations[index]

    def __len__(self):
        return len(self.annotations)

    def add(self,
            annotation_definition,
            object_id=None,
            frame_num=None,
            end_frame_num=None,
            start_time=None,
            end_time=None,
            automated=True,
            fixed=True,
            object_visible=True,
            metadata=None,
            parent_id=None,
            prompt_id=None,
            model_info=None):
        """
        Add annotations to collection

        :param annotation_definition: dl.Polygon, dl.Segmentation, dl.Point, dl.Box etc.
        :param object_id: Object id (any id given by user). If video - must input to match annotations between frames
        :param frame_num: video only, number of frame
        :param end_frame_num: video only, the end frame of the annotation
        :param start_time: video only, start time of the annotation
        :param end_time: video only, end time of the annotation
        :param automated:
        :param fixed: video only, mark frame as fixed
        :param object_visible: video only, does the annotated object is visible
        :param metadata: optional, metadata dictionary for annotation
        :param parent_id: set a parent for this annotation (parent annotation ID)
        :param prompt_id: Connect the annotation with a specific prompt in a dl.PromptItem
        :param model_info: optional - set model on annotation {'confidence': 0, # [Mandatory], (Float between 0-1)
                                                               'name': '', # [Optional], ('name' refers to 'model_name')
                                                               'model_id': ''} # [Optional]
        :return:
        """
        if model_info is not None:
            if not isinstance(model_info, dict) or 'confidence' not in model_info:
                raise ValueError('"model_info" must be a dict with key: "confidence"')
            if metadata is None:
                metadata = dict()
            if 'user' not in metadata:
                metadata['user'] = dict()
            confidence = float(model_info['confidence'])
            if confidence < 0 or confidence > 1:
                raise ValueError('"confidence" must be a float between 0 and 1')
            metadata['user']['model'] = {'confidence': confidence,
                                         'name': model_info.get('name'),
                                         'model_id': model_info.get('model_id'),
                                         }
        if prompt_id is not None:
            if metadata is None:
                metadata = dict()
            if 'system' not in metadata:
                metadata['system'] = dict()
            metadata['system']['promptId'] = prompt_id

        # to support list of definitions with same parameters
        if not isinstance(annotation_definition, list):
            annotation_definition = [annotation_definition]

        for single_definition in annotation_definition:
            annotation = entities.Annotation.new(item=self.item,
                                                 annotation_definition=single_definition,
                                                 frame_num=frame_num,
                                                 automated=automated,
                                                 metadata=metadata,
                                                 object_id=object_id,
                                                 parent_id=parent_id,
                                                 start_time=start_time,
                                                 end_time=end_time)
            #  add frame if exists
            if (frame_num is not None or start_time is not None) and (
                    self.item is None or 'audio' not in self.item.metadata.get('system').get('mimetype', '')):
                if object_id is None:
                    raise ValueError('Video Annotation must have object_id.')
                else:
                    if isinstance(object_id, int):
                        object_id = '{}'.format(object_id)
                    elif not isinstance(object_id, str) or not object_id.isnumeric():
                        raise ValueError('Object id must be an int or a string containing only numbers.')
                # find matching element_id
                matched_ind = [i_annotation
                               for i_annotation, annotation in enumerate(self.annotations)
                               if annotation.object_id == object_id]
                if len(matched_ind) == 0:
                    # no matching object id found - create new one
                    self.annotations.append(annotation)
                    matched_ind = len(self.annotations) - 1
                elif len(matched_ind) == 1:
                    matched_ind = matched_ind[0]
                else:
                    raise PlatformException(error='400',
                                            message='more than one annotation with same object id: {}'.format(
                                                object_id))

                self.annotations[matched_ind].add_frames(annotation_definition=single_definition,
                                                         frame_num=frame_num,
                                                         end_frame_num=end_frame_num,
                                                         start_time=start_time,
                                                         end_time=end_time,
                                                         fixed=fixed,
                                                         object_visible=object_visible)
            else:
                # add new annotation to list
                self.annotations.append(annotation)

    ############
    # Plotting #
    ############
    def show(self,
             image=None,
             thickness=None,
             with_text=False,
             height=None,
             width=None,
             annotation_format: entities.ViewAnnotationOptions = entities.ViewAnnotationOptions.MASK,
             label_instance_dict=None,
             color=None,
             alpha=1.,
             frame_num=None):
        """
        Show annotations according to annotation_format

        **Prerequisites**: Any user can upload annotations.

        :param ndarray image: empty or image to draw on
        :param int height: height
        :param int width: width
        :param int thickness: line thickness
        :param bool with_text: add label to annotation
        :param dl.ViewAnnotationOptions annotation_format: how to show thw annotations. options: list(dl.ViewAnnotationOptions)
        :param dict label_instance_dict: instance label map {'Label': 1, 'More': 2}
        :param tuple color: optional - color tuple
        :param float alpha: opacity value [0 1], default 1
        :param int frame_num: for video annotation, show specific frame
        :return: ndarray of the annotations

        **Example**:

        .. code-block:: python

            image = builder.show(image='ndarray',
                        thickness=1,
                        annotation_format=dl.VIEW_ANNOTATION_OPTIONS_MASK)
        """
        # if 'video' in self.item.mimetype and (annotation_format != 'json' or annotation_format != ['json']):
        #     raise PlatformException('400', 'Cannot show mask or instance of video item')
        # height/weight
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        # split the annotations to binary and not binary to put the binaries first
        segment_annotations = list()
        rest_annotations = list()
        for annotation in self.annotations:
            if annotation.type == 'binary':
                segment_annotations.append(annotation)
            else:
                rest_annotations.append(annotation)
        all_annotations = segment_annotations + rest_annotations
        # gor over all annotations and put the id where the annotations are
        for annotation in all_annotations:
            # get the mask of the annotation
            image = annotation.show(thickness=thickness,
                                    with_text=with_text,
                                    height=height,
                                    width=width,
                                    label_instance_dict=label_instance_dict,
                                    annotation_format=annotation_format,
                                    image=image,
                                    alpha=alpha,
                                    color=color,
                                    frame_num=frame_num)
        return image

    def _video_maker(self,
                     input_filepath,
                     output_filepath,
                     thickness=1,
                     annotation_format=entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE,
                     with_text=False,
                     alpha=1.):
        """
        create a video from frames

        :param input_filepath: str - input file path
        :param output_filepath: str - out put file path
        :param thickness: int - thickness of the annotations
        :param annotation_format: str - ViewAnnotationOptions - annotations format
        :param with_text: bool - if True show the label in the output
        :param float alpha: opacity value [0 1], default 1

        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        is_color = True
        if annotation_format in [entities.ViewAnnotationOptions.INSTANCE,
                                 entities.ViewAnnotationOptions.OBJECT_ID]:
            is_color = False
        # read input video
        fps = self.item.system.get('fps', 0)
        height = self.item.system.get('height', 0)
        width = self.item.system.get('width', 0)
        nb_frames = int(self.item.system.get('ffmpeg', {}).get('nb_read_frames'))
        writer = cv2.VideoWriter(filename=output_filepath,
                                 fourcc=cv2.VideoWriter_fourcc(*"mp4v"),
                                 fps=fps,
                                 frameSize=(width, height),
                                 isColor=is_color)
        if input_filepath is not None and is_color:
            reader = cv2.VideoCapture(input_filepath)
        else:
            reader = None

        for frame_num in range(nb_frames):
            if reader is not None:
                ret, frame = reader.read()
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGBA)
            else:
                frame = None
            frame = self.show(image=frame,
                              annotation_format=annotation_format,
                              thickness=thickness,
                              alpha=alpha,
                              height=height,
                              width=width,
                              with_text=with_text,
                              frame_num=frame_num)
            if is_color:
                frame = cv2.cvtColor(frame, cv2.COLOR_RGBA2BGR)
            writer.write(frame)
        writer.release()
        if reader is not None:
            reader.release()

    @staticmethod
    def _set_flip_args(orientation):
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            logger.error(
                'Import Error! Cant import cv2. that make the exif orientation not supported')
            raise
        if orientation == 3:
            return cv2.ROTATE_180, cv2.ROTATE_180, 0
        elif orientation == 4:
            return cv2.ROTATE_180, cv2.ROTATE_180, 1
        elif orientation == 5:
            return cv2.ROTATE_90_CLOCKWISE, cv2.ROTATE_90_COUNTERCLOCKWISE, 1
        elif orientation == 6:
            return cv2.ROTATE_90_CLOCKWISE, cv2.ROTATE_90_COUNTERCLOCKWISE, 0
        elif orientation == 7:
            return cv2.ROTATE_90_COUNTERCLOCKWISE, cv2.ROTATE_90_CLOCKWISE, 1
        else:
            return cv2.ROTATE_90_COUNTERCLOCKWISE, cv2.ROTATE_90_CLOCKWISE, 0

    def download(self,
                 filepath,
                 img_filepath=None,
                 annotation_format: entities.ViewAnnotationOptions = entities.ViewAnnotationOptions.JSON,
                 height=None,
                 width=None,
                 thickness=1,
                 with_text=False,
                 orientation=0,
                 alpha=1):
        """
        Save annotations to file

        **Prerequisites**: Any user can upload annotations.

        :param str filepath: path to save annotation
        :param str img_filepath: img file path - needed for img_mask
        :param dl.ViewAnnotationOptions annotation_format: how to show thw annotations. options: list(dl.ViewAnnotationOptions)
        :param int height: height
        :param int width: width
        :param int thickness: thickness
        :param bool with_text: add a text to an image
        :param int orientation: the image orientation
        :param float alpha: opacity value [0 1], default 1
        :return: file path of the download annotation
        :rtype: str

        **Example**:

        .. code-block:: python

            filepath = builder.download(filepath='filepath', annotation_format=dl.ViewAnnotationOptions.MASK)
        """
        dir_name, ex = os.path.splitext(filepath)
        if annotation_format == entities.ViewAnnotationOptions.JSON:
            if not ex:
                filepath = os.path.join(dir_name, '{}.json'.format(os.path.splitext(self.item.name)[0]))
            _json = dict()
            if self.item is not None:
                _json = {'_id': self.item.id,
                         'filename': self.item.filename}
            annotations = list()
            for ann in self.annotations:
                annotations.append(ann.to_json())
            _json['annotations'] = annotations
            if self.item is not None:
                _json['metadata'] = self.item.metadata
            with open(filepath, 'w+') as f:
                json.dump(_json, f, indent=2)
        elif annotation_format in [entities.ViewAnnotationOptions.MASK,
                                   entities.ViewAnnotationOptions.INSTANCE,
                                   entities.ViewAnnotationOptions.OBJECT_ID,
                                   entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE]:
            if not ex:
                if 'video' in self.item.mimetype:
                    filepath = os.path.join(dir_name, '{}.mp4'.format(os.path.splitext(self.item.name)[0]))
                else:
                    filepath = os.path.join(dir_name, '{}.png'.format(os.path.splitext(self.item.name)[0]))
            image = None
            if 'video' in self.item.mimetype:
                if annotation_format == entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE:
                    if img_filepath is None:
                        img_filepath = self.item.download()
                    annotation_format = entities.ViewAnnotationOptions.MASK
                self._video_maker(input_filepath=img_filepath,
                                  output_filepath=filepath,
                                  thickness=thickness,
                                  annotation_format=annotation_format,
                                  with_text=with_text,
                                  alpha=alpha
                                  )
                return filepath
            if annotation_format == entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE:
                if img_filepath is None:
                    img_filepath = self.item.download()
                annotation_format = entities.ViewAnnotationOptions.MASK
                image = np.asarray(Image.open(img_filepath))
            if orientation in [3, 4, 5, 6, 7, 8]:
                try:
                    import cv2
                except (ImportError, ModuleNotFoundError):
                    logger.error(
                        'Import Error! Cant import cv2. that make the exif orientation not supported')
                    raise
                first_rotate, second_rotate, flip = self._set_flip_args(orientation=orientation)
                image = cv2.rotate(image, first_rotate)
                if flip:
                    image = np.flip(image, 1)
            mask = self.show(image=image,
                             thickness=thickness,
                             with_text=with_text,
                             height=height,
                             width=width,
                             alpha=alpha,
                             annotation_format=annotation_format)
            if orientation not in [3, 4, 5, 6, 7, 8]:
                img = Image.fromarray(mask.astype(np.uint8))
            else:
                img = Image.fromarray(cv2.rotate(mask.astype(np.uint8), second_rotate))
            img.save(filepath)
        elif annotation_format == entities.ViewAnnotationOptions.VTT:
            if not ex:
                filepath = '{}/{}.vtt'.format(dir_name, os.path.splitext(self.item.name)[0])
            annotations_dict = [{'start_time': annotation.start_time,
                                 'end_time': annotation.end_time,
                                 'text': annotation.coordinates['text']} for annotation in self.annotations if
                                annotation.type in ['subtitle']]
            sorted_by_start_time = sorted(annotations_dict, key=lambda i: i['start_time'])
            vtt = webvtt.WebVTT()
            for ann in sorted_by_start_time:
                s = str(datetime.timedelta(seconds=ann['start_time']))
                if len(s.split('.')) == 1:
                    s += '.000'
                e = str(datetime.timedelta(seconds=ann['end_time']))
                if len(e.split('.')) == 1:
                    e += '.000'
                caption = webvtt.Caption(
                    '{}'.format(s),
                    '{}'.format(e),
                    '{}'.format(ann['text'])
                )
                vtt.captions.append(caption)
            vtt.save(filepath)
        else:
            raise PlatformException(error="400", message="Unknown annotation option: {}".format(annotation_format))
        return filepath

    ############
    # Platform #
    ############
    def update(self, system_metadata=True):
        """
        Update an existing annotation in host.

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or *developer* or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :param system_metadata: True, if you want to change metadata system
        :return: Annotation object
        :rtype: dtlpy.entities.annotation.Annotation

        **Example**:

        .. code-block:: python

            annotation = builder.update()
        """
        if self.item is None:
            raise PlatformException('400', 'missing item to perform platform update')
        return self.item.annotations.update(annotations=self.annotations, system_metadata=system_metadata)

    def delete(self):
        """
        Remove an annotation from item

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or *developer* or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            is_deleted = builder.delete()
        """
        if self.item is None:
            raise PlatformException('400', 'missing item to perform platform delete')
        return [self.item.annotations.delete(annotation_id=annotation.id) for annotation in self.annotations]

    def upload(self):
        """
        Create a new annotation in host

        **Prerequisites**: Any user can upload annotations.

        :return: Annotation entity
        :rtype: dtlpy.entities.annotation.Annotation

        **Example**:

        .. code-block:: python

            annotation = builder.upload()
        """
        if self.item is None:
            raise PlatformException('400', 'missing item to perform platform upload')
        return self.item.annotations.upload(self.annotations)

    @staticmethod
    def _json_to_annotation(item: entities.Item, w_json: dict, is_video=None, fps=25, item_metadata=None,
                            client_api=None):
        try:
            annotation = entities.Annotation.from_json(_json=w_json,
                                                       fps=fps,
                                                       item_metadata=item_metadata,
                                                       is_video=is_video,
                                                       item=item,
                                                       client_api=client_api)
            status = True
        except Exception:
            annotation = traceback.format_exc()
            status = False
        return status, annotation

    @classmethod
    def from_json(cls, _json: list, item=None, is_video=None, fps=25, height=None, width=None,
                  client_api=None, is_audio=None) -> 'AnnotationCollection':
        """
        Create an annotation collection object from platform json

        :param dict _json: platform json
        :param dtlpy.entities.item.Item item: item
        :param client_api: ApiClient entity
        :param bool is_video: whether the item is video
        :param fps: frame rate of the video
        :param float height: height
        :param float width: width
        :param bool is_audio: is audio
        :return: annotation object
        :rtype: dtlpy.entities.annotation.Annotation
        """
        if item is None:
            if isinstance(_json, dict):
                metadata = _json.get('metadata', dict())
                system_metadata = metadata.get('system', dict())
                if is_video is None:
                    if 'mimetype' in system_metadata:
                        is_video = 'video' in system_metadata['mimetype']
                        is_audio = 'audio' in system_metadata['mimetype']
                    elif 'filename' in _json:
                        ext = os.path.splitext(_json['filename'])[-1]
                        try:
                            is_video = 'video' in mimetypes.types_map[ext.lower()]
                            is_audio = 'is_audio' in mimetypes.types_map[ext.lower()]
                        except Exception:
                            logger.info("Unknown annotation's item type. Default item type is set to: image")
                    else:
                        logger.info("Unknown annotation's item type. Default item type is set to: image")
                if is_video:
                    fps = system_metadata.get('fps', fps)
                    ffmpeg_info = system_metadata.get('ffmpeg', dict())
                    height = ffmpeg_info.get('height', None)
                    width = ffmpeg_info.get('width', None)
                elif is_audio:
                    fps = system_metadata.get('fps', 1000)
                    height = system_metadata.get('height', None)
                    width = system_metadata.get('width', None)
                else:
                    fps = 0
                    height = system_metadata.get('height', None)
                    width = system_metadata.get('width', None)
        else:
            if client_api is None:
                client_api = item._client_api
            fps = item.fps
            height = item.height
            width = item.width

        item_metadata = {
            'fps': fps,
            'height': height,
            'width': width
        }

        if 'annotations' in _json:
            _json = _json['annotations']

        results = list([None for _ in range(len(_json))])
        for i_json, single_json in enumerate(_json):
            results[i_json] = cls._json_to_annotation(item=item,
                                                      fps=fps,
                                                      item_metadata=item_metadata,
                                                      is_video=is_video,
                                                      w_json=single_json,
                                                      client_api=client_api)
        # log errors
        _ = [logger.warning(j[1]) for j in results if j[0] is False]

        # return good jobs
        annotations = [j[1] for j in results if j[0] is True]

        # sort
        if is_video:
            annotations.sort(key=lambda x: x.start_frame)
        else:
            annotations.sort(key=lambda x: x.label)

        return cls(annotations=miscellaneous.List(annotations), item=item)

    @classmethod
    def from_json_file(cls, filepath, item=None):
        with open(filepath, 'r', encoding='utf-8') as f:
            data = json.load(f)
        return cls.from_json(_json=data, item=item)

    def from_vtt_file(self, filepath):
        """
        convert annotation from vtt format

        :param str filepath: path to the file
        """
        for caption in webvtt.read(filepath):
            h, m, s = caption.start.split(':')
            start_time = datetime.timedelta(hours=float(h), minutes=float(m), seconds=float(s)).total_seconds()
            h, m, s = caption.end.split(':')
            end_time = datetime.timedelta(hours=float(h), minutes=float(m), seconds=float(s)).total_seconds()
            annotation_definition = entities.Subtitle(text=caption.text, label='Text')
            annotation = entities.Annotation.new(
                annotation_definition=annotation_definition,
                item=self.item,
                start_time=start_time)

            annotation.add_frames(annotation_definition=annotation_definition,
                                  start_time=start_time,
                                  end_time=end_time)

            self.annotations.append(annotation)

    def from_instance_mask(self, mask, instance_map=None):
        """
        convert annotation from instance mask format

        :param mask: the mask annotation
        :param instance_map: labels
        """
        if instance_map is None:
            instance_map = self.item.dataset.instance_map
        # go over all instance ids
        for label, instance_id in instance_map.items():
            # find a binary mask per instance
            class_mask = instance_id == mask
            if not np.any(class_mask):
                continue
            # add the binary mask to the annotation builder
            self.add(annotation_definition=entities.Segmentation(geo=class_mask, label=label))

    def to_json(self):
        """
        Convert annotation object to a platform json representation

        :return: platform json
        :rtype: dict
        """
        if self.item is None:
            item_id = None
            item_name = None
        else:
            item_id = self.item.id
            item_name = self.item.filename

        _json = {
            "_id": item_id,
            "filename": item_name,
            'annotations': [annotation.to_json() for annotation in self.annotations]
        }

        return _json

    def print(self, to_return=False, columns=None):
        """

        :param to_return:
        :param columns:
        """
        return miscellaneous.List(self.annotations).print(to_return=to_return, columns=columns)

    #########################
    # For video annotations #
    #########################
    def get_frame(self, frame_num):
        """
        Get frame

        :param int frame_num: frame num
        :return: AnnotationCollection
        """
        frame_collection = AnnotationCollection(item=self.item)
        for annotation in self.annotations:
            if frame_num in annotation.frames:
                annotation.set_frame(frame=frame_num)
                frame_collection.annotations.append(annotation)
        return frame_collection

    def video_player(self):
        from ..utilities.videos.video_player import VideoPlayer
        _ = VideoPlayer(dataset_id=self.item.dataset.id,
                        item_id=self.item.id)



import traceback
from enum import Enum
from typing import List, Optional, Dict
from ..services.api_client import ApiClient
from .. import repositories


class ClusterProvider(str, Enum):
    GCP = 'gcp'
    AWS = 'aws'
    AZURE = 'azure'
    HPC = 'hpc'
    LOCAL = 'local'
    RANCHER_K3S = 'rancher-k3s'
    RANCHER_RKE = 'rancher-rke'
    OPENSHIFT = 'openshift'


class ComputeType(str, Enum):
    KUBERNETES = "kubernetes"


class ComputeStatus(str, Enum):
    READY = "ready"
    INITIALIZING = "initializing"
    PAUSE = "pause"
    FAILED = "failed"
    VALIDATING = "validating"


class ComputeConsumptionMethod(str, Enum):
    MQ = "MQ",
    API = "API"


class ComputeSettings:
    def __init__(self, default_namespace: str, consumption_method: ComputeConsumptionMethod):
        self.consumption_method = consumption_method
        self.default_namespace = default_namespace

    @classmethod
    def from_json(cls, _json):
        return cls(
            default_namespace=_json.get('defaultNamespace'),
            consumption_method=_json.get('consumptionMethod')
        )

    def to_json(self):
        return {
            'defaultNamespace': self.default_namespace,
            'consumptionMethod': self.consumption_method
        }


class Toleration:
    def __init__(self, effect: str, key: str, operator: str, value: str):
        self.effect = effect
        self.key = key
        self.operator = operator
        self.value = value

    @classmethod
    def from_json(cls, _json):
        return cls(
            effect=_json.get('effect'),
            key=_json.get('key'),
            operator=_json.get('operator'),
            value=_json.get('value')
        )

    def to_json(self):
        return {
            'effect': self.effect,
            'key': self.key,
            'operator': self.operator,
            'value': self.value
        }


class DeploymentResource:
    def __init__(self, gpu: int, cpu: int, memory: str):
        self.gpu = gpu
        self.cpu = cpu
        self.memory = memory

    @classmethod
    def from_json(cls, _json):
        return cls(
            gpu=_json.get('gpu', None),
            cpu=_json.get('cpu', None),
            memory=_json.get('memory', None)
        )

    def to_json(self):
        _json = {}
        if self.gpu is not None:
            _json['gpu'] = self.gpu
        if self.cpu is not None:
            _json['cpu'] = self.cpu
        if self.memory is not None:
            _json['memory'] = self.memory


class DeploymentResources:
    def __init__(self, request: DeploymentResource, limit: DeploymentResource):
        self.request = request
        self.limit = limit

    @classmethod
    def from_json(cls, _json):
        return cls(
            request=DeploymentResource.from_json(_json.get('request') or dict()),
            limit=DeploymentResource.from_json(_json.get('limit') or dict())
        )

    def to_json(self):
        return {
            'request': self.request.to_json(),
            'limit': self.limit.to_json()
        }


class NodePool:
    def __init__(
            self,
            name: str,
            is_dl_type_default: bool,
            dl_types: Optional[List[str]] = None,
            tolerations: Optional[List[Toleration]] = None,
            description: str = "",
            node_selector: str = "",
            preemptible: bool = False,
            deployment_resources: DeploymentResources = None
    ):
        self.name = name
        self.is_dl_type_default = is_dl_type_default
        self.dl_types = dl_types
        self.tolerations = tolerations if tolerations is not None else []
        self.description = description
        self.node_selector = node_selector
        self.preemptible = preemptible
        self.deployment_resources = deployment_resources

    @classmethod
    def from_json(cls, _json):
        node_pool = cls(
            name=_json.get('name'),
            is_dl_type_default=_json.get('isDlTypeDefault'),
            dl_types=_json.get('dlTypes'),
            description=_json.get('description'),
            node_selector=_json.get('nodeSelector'),
            preemptible=_json.get('preemptible'),
            deployment_resources=DeploymentResources.from_json(_json.get('deploymentResources', dict())),
            tolerations=[Toleration.from_json(t) for t in _json.get('tolerations', list())]
        )

        return node_pool

    def to_json(self):
        _json = {
            'name': self.name,
            'isDlTypeDefault': self.is_dl_type_default,
            'description': self.description,
            'nodeSelector': self.node_selector,
            'preemptible': self.preemptible,
            'deploymentResources': self.deployment_resources.to_json(),
            'tolerations': [t.to_json() for t in self.tolerations]
        }

        if self.dl_types is not None:
            _json['dlTypes'] = self.dl_types

        return _json


class AuthenticationIntegration:
    def __init__(self, id: str, type: str):
        self.id = id
        self.type = type

    @classmethod
    def from_json(cls, _json):
        return cls(
            id=_json.get('id'),
            type=_json.get('type')
        )

    def to_json(self):
        return {
            'id': self.id,
            'type': self.type
        }


class Authentication:
    def __init__(self, integration: AuthenticationIntegration):
        self.integration = integration

    @classmethod
    def from_json(cls, _json):
        return cls(
            integration=AuthenticationIntegration.from_json(_json.get('integration', dict()))
        )

    def to_json(self):
        return {
            'integration': self.integration.to_json()
        }


class ComputeCluster:
    def __init__(
            self,
            name: str,
            endpoint: str,
            kubernetes_version: str,
            provider: ClusterProvider,
            node_pools: Optional[List[NodePool]] = None,
            metadata: Optional[Dict] = None,
            authentication: Optional[Authentication] = None,
            plugins: Optional[dict] = None,
            deployment_configuration: Optional[Dict] = None
    ):
        self.name = name
        self.endpoint = endpoint
        self.kubernetes_version = kubernetes_version
        self.provider = provider
        self.node_pools = node_pools if node_pools is not None else []
        self.metadata = metadata if metadata is not None else {}
        self.authentication = authentication if authentication is not None else Authentication(
            AuthenticationIntegration("", ""))
        self.plugins = plugins
        self.deployment_configuration = deployment_configuration if deployment_configuration is not None else {}



    @classmethod
    def from_json(cls, _json):
        return cls(
            name=_json.get('name'),
            endpoint=_json.get('endpoint'),
            kubernetes_version=_json.get('kubernetesVersion'),
            provider=ClusterProvider(_json.get('provider')),
            node_pools=[NodePool.from_json(np) for np in _json.get('nodePools', list())],
            metadata=_json.get('metadata'),
            authentication=Authentication.from_json(_json.get('authentication', dict())),
            plugins=_json.get('plugins'),
            deployment_configuration=_json.get('deploymentConfiguration'),
        )

    def to_json(self):
        return {
            'name': self.name,
            'endpoint': self.endpoint,
            'kubernetesVersion': self.kubernetes_version,
            'provider': self.provider.value,
            'nodePools': [np.to_json() for np in self.node_pools],
            'metadata': self.metadata,
            'authentication': self.authentication.to_json(),
            'plugins': self.plugins,
            'deploymentConfiguration':  self.deployment_configuration
        }

    @classmethod
    def from_setup_json(cls, devops_output, integration):
        node_pools = [NodePool.from_json(n) for n in devops_output['config']['nodePools']]
        return cls(
            name=devops_output['config']['name'],
            endpoint=devops_output['config']['endpoint'],
            kubernetes_version=devops_output['config']['kubernetesVersion'],
            provider=ClusterProvider(devops_output['config']['provider']),
            node_pools=node_pools,
            metadata={},
            authentication=Authentication(AuthenticationIntegration(integration.id, integration.type)),
            plugins=devops_output['config'].get('plugins'),
            deployment_configuration=devops_output['config'].get('deploymentConfiguration', {})
        )


class ComputeContext:
    def __init__(self, labels: List[str], org: str, project: Optional[str] = None):
        self.labels = labels
        self.org = org
        self.project = project

    @classmethod
    def from_json(cls, _json):
        return cls(
            labels=_json.get('labels', list()),
            org=_json.get('org'),
            project=_json.get('project')
        )

    def to_json(self):
        return {
            'labels': self.labels,
            'org': self.org,
            'project': self.project
        }


class Compute:
    def __init__(
            self,
            id: str,
            name: str,
            context: ComputeContext,
            client_api: ApiClient,
            shared_contexts: Optional[List[ComputeContext]] = None,
            global_: Optional[bool] = None,
            status: ComputeStatus = ComputeStatus.INITIALIZING,
            type: ComputeType = ComputeType.KUBERNETES,
            features: Optional[Dict] = None,
            metadata: Optional[Dict] = None,
            settings: Optional[ComputeSettings] = None,
            url: Optional[str] = None
    ):
        self.id = id
        self.name = name
        self.context = context
        self.shared_contexts = shared_contexts if shared_contexts is not None else []
        self.global_ = global_
        self.status = status
        self.type = type
        self.features = features if features is not None else dict()
        self.metadata = metadata if metadata is not None else dict()
        self._client_api = client_api
        self._computes = None
        self._serviceDrivers = None
        self.settings = settings
        self.url = url

    @property
    def computes(self):
        if self._computes is None:
            self._computes = repositories.Computes(client_api=self._client_api)
        return self._computes

    def delete(self, skip_destroy: bool = False):
        return self.computes.delete(compute_id=self.id, skip_destroy=skip_destroy)

    def update(self):
        return self.computes.update(compute=self)

    @staticmethod
    def _protected_from_json(_json: dict, client_api: ApiClient):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform json
        :param client_api: ApiClient entity
        :return:
        """
        try:
            compute = Compute.from_json(_json=_json,
                                        client_api=client_api)
            status = True
        except Exception:
            compute = traceback.format_exc()
            status = False
        return status, compute

    @classmethod
    def from_json(cls, _json, client_api: ApiClient):
        return cls(
            id=_json.get('id'),
            name=_json.get('name'),
            context=ComputeContext.from_json(_json.get('context', dict())),
            shared_contexts=[ComputeContext.from_json(sc) for sc in _json.get('sharedContexts', list())],
            global_=_json.get('global'),
            status=ComputeStatus(_json.get('status')),
            type=ComputeType(_json.get('type')),
            features=_json.get('features'),
            client_api=client_api,
            metadata=_json.get('metadata'),
            settings=ComputeSettings.from_json(_json.get('settings', dict())) if _json.get('settings') else None,
            url=_json.get('url'),
        )

    def to_json(self):
        return {
            'id': self.id,
            'name': self.name,
            'context': self.context.to_json(),
            'sharedContexts': [sc.to_json() for sc in self.shared_contexts],
            'global': self.global_,
            'status': self.status.value,
            'type': self.type.value,
            'features': self.features,
            'metadata': self.metadata,
            'settings': self.settings.to_json() if isinstance(self.settings, ComputeSettings) else self.settings,
            'url': self.url
        }


class KubernetesCompute(Compute):
    def __init__(
            self,
            id: str,
            name: str,
            context: ComputeContext,
            cluster: ComputeCluster,
            shared_contexts: Optional[List[ComputeContext]] = None,
            global_: Optional[bool] = None,
            status: ComputeStatus = ComputeStatus.INITIALIZING,
            type: ComputeType = ComputeType.KUBERNETES,
            features: Optional[Dict] = None,
            metadata: Optional[Dict] = None,
            client_api: ApiClient = None,
            settings: Optional[ComputeSettings] = None,
            url: Optional[str] = None
    ):
        super().__init__(id=id, context=context, shared_contexts=shared_contexts, global_=global_, status=status,
                         type=type, features=features, metadata=metadata, client_api=client_api, settings=settings,
                         name=name, url=url)
        self.cluster = cluster

    @classmethod
    def from_json(cls, _json, client_api: ApiClient):
        return cls(
            id=_json.get('id'),
            name=_json.get('name'),
            context=ComputeContext.from_json(_json.get('context', dict())),
            cluster=ComputeCluster.from_json(_json.get('cluster', dict())),
            shared_contexts=[ComputeContext.from_json(sc) for sc in _json.get('sharedContexts', list())],
            global_=_json.get('global'),
            status=ComputeStatus(_json.get('status')),
            type=ComputeType(_json.get('type')),
            features=_json.get('features'),
            metadata=_json.get('metadata'),
            client_api=client_api,
            settings=ComputeSettings.from_json(_json.get('settings', dict())) if _json.get('settings') else None,
            url=_json.get('url')
        )

    def to_json(self):
        return {
            'id': self.id,
            'name': self.name,
            'context': self.context.to_json(),
            'cluster': self.cluster.to_json(),
            'sharedContexts': [sc.to_json() for sc in self.shared_contexts],
            'global': self.global_,
            'status': self.status.value,
            'type': self.type.value,
            'features': self.features,
            'metadata': self.metadata,
            'settings': self.settings.to_json() if isinstance(self.settings, ComputeSettings) else self.settings,
            'url': self.url
        }



import base64
import datetime
import json

from dtlpy import miscellaneous

from ..services.api_client import ApiClient
from .. import exceptions, entities, repositories
from typing import List, Optional, Dict
from ..entities import ComputeCluster, ComputeContext, ComputeType
from ..entities.integration import IntegrationType
import logging
from urllib.parse import urlparse, urlencode, parse_qs, urlunparse

logger = logging.getLogger(name='dtlpy')


class Computes:

    def __init__(self, client_api: ApiClient):
        self._client_api = client_api
        self._base_url = '/compute'
        self._commands = None
        self._projects = None
        self._organizations = None
        self.log_cache = dict()

    @property
    def commands(self) -> repositories.Commands:
        if self._commands is None:
            self._commands = repositories.Commands(client_api=self._client_api)
        return self._commands

    @property
    def projects(self):
        if self._projects is None:
            self._projects = repositories.Projects(client_api=self._client_api)
        return self._projects

    @property
    def organizations(self):
        if self._organizations is None:
            self._organizations = repositories.Organizations(client_api=self._client_api)
        return self._organizations

    def create(
            self,
            name: str,
            context: entities.ComputeContext,
            shared_contexts: Optional[List[entities.ComputeContext]],
            cluster: entities.ComputeCluster,
            type: entities.ComputeType = entities.ComputeType.KUBERNETES,
            is_global: Optional[bool] = False,
            features: Optional[Dict] = None,
            wait=True,
            status: entities.ComputeStatus = None,
            settings: entities.ComputeSettings = None,
            metadata: dict = None,
            deployment_configuration: dict = None
    ):
        """
        Create a new compute

        :param name: Compute name
        :param context: Compute context
        :param shared_contexts: Shared contexts
        :param cluster: Compute cluster
        :param type: Compute type
        :param is_global: Is global
        :param features: Features
        :param wait: Wait for compute creation
        :param status: Compute status
        :param settings: Compute settings
        :param metadata: Compute metadata
        :param deployment_configuration: Compute deployment Configuration
        :return: Compute
        :rtype: dl.entities.compute.Compute
        """
        if metadata is None:
            metadata = {}
        shared_contexts_json = []
        for shared_context in shared_contexts:
            src_json = shared_context.to_json() if isinstance(shared_context,
                                                              entities.ComputeContext) else shared_context
            shared_contexts_json.append(src_json)
        payload = {
            'name': name,
            'context': context.to_json(),
            'type': type.value,
            'global': is_global,
            'features': features,
            'sharedContexts': shared_contexts_json,
            'cluster': cluster.to_json(),
            'status': status,
            "settings": settings.to_json() if isinstance(settings, entities.ComputeSettings) else settings,
            "metadata": metadata,
            "deploymentConfiguration": deployment_configuration
        }

        # request
        success, response = self._client_api.gen_request(
            req_type='post',
            path=self._base_url,
            json_req=payload
        )

        if not success:
            raise exceptions.PlatformException(response)

        compute = self._build_compute_by_type(response.json())

        if wait:
            command_id = compute.metadata.get('system', {}).get('commands', {}).get('create', None)
            if command_id is not None:
                command = self.commands.get(command_id=command_id, url='api/v1/commands/faas/{}'.format(command_id))
                try:
                    callback = self.__get_log_compute_progress_callback(compute.id)
                    command.wait(iteration_callback=callback)
                    callback()
                except Exception as e:
                    self.log_cache.pop(compute.id, None)
                    raise e
                compute = self.get(compute_id=compute.id)

        return compute

    def _build_compute_by_type(self, _json):
        if _json.get('type') == 'kubernetes':
            compute = entities.KubernetesCompute.from_json(
                _json=_json,
                client_api=self._client_api
            )
        else:
            compute = entities.Compute.from_json(
                _json=_json,
                client_api=self._client_api
            )
        return compute

    def __get_log_compute_progress_callback(self, compute_id: str, is_destroy=False):
        def func():
            compute = self.get(compute_id=compute_id, archived=True)
            log_type = 'bootstrap'
            validation_progress = None
            validation_logs = None
            if is_destroy is True:
                log_type = 'destroy'
            else:
                validation_progress = compute.metadata.get('system', {}).get('validation', {}).get('progress', None)
                validation_logs = compute.metadata.get('system', {}).get('validation', {}).get('logs', None)
            bootstrap_progress = compute.metadata.get('system', {}).get(log_type, {}).get('progress', None)
            bootstrap_logs = compute.metadata.get('system', {}).get(log_type, {}).get('logs', None)

            if bootstrap_progress is not None:
                if 'bootstrap' not in self.log_cache.get(compute_id, {}):
                    logger.info(f"{log_type} in progress:")
                last_index = len(self.log_cache.get(compute_id, {}).get(log_type, []))
                new_logs = bootstrap_logs[last_index:]
                if new_logs:
                    for log in new_logs:
                        logger.info(log)
                    logger.info(f'{log_type} progress: {int(bootstrap_progress)}%')
                    if compute_id not in self.log_cache:
                        self.log_cache[compute_id] = {}
                    self.log_cache[compute_id][log_type] = bootstrap_logs
            if bootstrap_progress in [100, None] and validation_progress is not None:
                if 'validation' not in self.log_cache.get(compute_id, {}):
                    logger.info(f"Validating created compute:")
                last_index = len(self.log_cache.get(compute_id, {}).get('validation', []))
                new_logs = validation_logs[last_index:]
                if new_logs:
                    for log in new_logs:
                        logger.info(log)
                    logger.info(f'Validation progress: {int(validation_progress)}%')
                    if compute_id not in self.log_cache:
                        self.log_cache[compute_id] = {}
                    self.log_cache[compute_id]['validation'] = validation_logs

        return func

    def get(self, compute_id: str, archived=False):
        """
        Get a compute

        :param compute_id: Compute ID
        :param archived: Archived
        :return: Compute
        :rtype: dl.entities.compute.Compute
        """
        url_path = self._base_url + '/{}'.format(compute_id)
        params_to_add = {"archived": "true" if archived else "false"}
        parsed_url = urlparse(url_path)
        query_dict = parse_qs(parsed_url.query)
        query_dict.update(params_to_add)
        new_query = urlencode(query_dict, doseq=True)
        url_path = urlunparse(parsed_url._replace(query=new_query))
        # request
        success, response = self._client_api.gen_request(
            req_type='get',
            path=url_path
        )

        if not success:
            raise exceptions.PlatformException(response)

        compute = self._build_compute_by_type(response.json())

        return compute

    def update(self, compute: entities.Compute):
        """
        Update a compute

        :param compute: Compute
        :return: Compute
        :rtype: dl.entities.compute.Compute
        """

        # request
        success, response = self._client_api.gen_request(
            req_type='patch',
            path=self._base_url + '/{}'.format(compute.id),
            json_req=compute.to_json()
        )

        if not success:
            raise exceptions.PlatformException(response)

        compute = self._build_compute_by_type(response.json())

        return compute

    def delete(self, compute_id: str, skip_destroy: bool = False, wait: bool = True):
        """
        Delete a compute

        :param compute_id: compute ID
        :param skip_destroy: bool
        :param bool wait: Wait for deletion
        """
        url_path = self._base_url + '/{}'.format(compute_id)
        params_to_add = {"skipDestroy": "true" if skip_destroy else "false"}
        parsed_url = urlparse(url_path)
        query_dict = parse_qs(parsed_url.query)
        query_dict.update(params_to_add)
        new_query = urlencode(query_dict, doseq=True)
        url_path = urlunparse(parsed_url._replace(query=new_query))
        # request
        success, response = self._client_api.gen_request(
            req_type='delete',
            path=url_path
        )

        if not success:
            raise exceptions.PlatformException(response)
        if skip_destroy is not True and wait is True:
            command_response = response.json()
            command_id = command_response['id']
            command = self.commands.get(command_id, url='api/v1/commands/faas/{}'.format(command_id))
            try:
                callback = self.__get_log_compute_progress_callback(compute_id, is_destroy=True)
                command.wait(iteration_callback=callback)
                callback()
            except Exception as e:
                self.log_cache.pop(command_id, None)
                raise e

        return True

    def validate(self, compute_id: str, wait: bool = True):
        """
        Validate a compute

        :param str compute_id: Compute ID
        :param bool wait: Wait for validation
        :return: Compute
        :rtype: dl.entities.compute.Compute
        """

        # request
        success, response = self._client_api.gen_request(
            req_type='post',
            path=self._base_url + '/{}/validate'.format(compute_id)
        )

        if not success:
            raise exceptions.PlatformException(response)

        compute = self._build_compute_by_type(response.json())

        if wait:
            command_id = compute.metadata.get('system', {}).get('commands', {}).get('validate', None)
            if command_id is not None:
                command = self.commands.get(command_id=command_id, url='api/v1/commands/faas/{}'.format(command_id))
                try:
                    callback = self.__get_log_compute_progress_callback(compute.id)
                    command.wait(iteration_callback=callback)
                    callback()
                except Exception as e:
                    self.log_cache.pop(compute.id, None)
                    raise e
                compute = self.get(compute_id=compute.id)

        return compute

    def list_global(self):
        """
        List computes

        :return: List of computes
        :rtype: list[str]
        """

        # request
        success, response = self._client_api.gen_request(
            req_type='get',
            path=self._base_url + '/globals',
        )

        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    @staticmethod
    def read_file(file_path):
        try:
            with open(file_path, 'r') as file:
                content = file.read()
            return content
        except FileNotFoundError:
            print(f"The file at {file_path} was not found.")
        except IOError:
            print(f"An error occurred while reading the file at {file_path}.")

    def decode_and_parse_input(self, file_path):
        """Decode a base64 encoded string from file a and parse it as JSON."""
        decoded_bytes = base64.b64decode(self.read_file(file_path))
        return json.loads(decoded_bytes)

    @staticmethod
    def create_integration(org, name, auth_data):
        """Create a new key-value integration within the specified project."""
        return org.integrations.create(
            integrations_type=IntegrationType.KEY_VALUE,
            name=name,
            options={
                'key': name,
                'value': json.dumps(auth_data)
            }
        )

    def setup_compute_cluster(self, config, integration, org_id, project=None, is_global=False):
        """Set up a compute cluster using the provided configuration and integration."""
        cluster = ComputeCluster.from_setup_json(config, integration)
        project_id = None
        if project is not None:
            project_id = project.id
        compute = self.create(
            config['config']['name'],
            ComputeContext([], org_id, project_id),
            [],
            cluster,
            ComputeType.KUBERNETES,
            status=config['config'].get('status', None),
            settings=config['config'].get('settings', None),
            deployment_configuration=config['config'].get('deploymentConfiguration', {}),
            metadata=config['config'].get('metadata', None), is_global=is_global)

        return compute

    def create_from_config_file(self, config_file_path, org_id, project_name: Optional[str] = None, is_global=False):
        config = self.decode_and_parse_input(config_file_path)
        project = None
        if project_name is not None:
            project = self.projects.get(project_name=project_name)
        org = self.organizations.get(organization_id=org_id)
        integration_name = ('cluster_integration_test_' + datetime.datetime.now().isoformat().split('.')[0]
                            .replace(':', '_'))
        integration = self.create_integration(org, integration_name, config['authentication'])
        compute = self.setup_compute_cluster(config, integration, org_id, project, is_global=is_global)
        return compute

    def _list(self, filters: entities.Filters):
        url = self._base_url + '/query'
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Compute]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.submit(entities.Compute._protected_from_json,
                                       **{'client_api': self._client_api,
                                          '_json': item})
        results = [j.result() for j in jobs]
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        items = miscellaneous.List([r[1] for r in results if r[0] is True])
        return items

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List all services drivers

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            services = dl.service_drivers.list()
        """
        # default filters
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.COMPUTE)

        if filters.resource != entities.FiltersResource.COMPUTE:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.COMPUTE. Got: {!r}'.format(
                    filters.resource))

        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException('400', 'Unknown filters type')

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged



def main():
    import dtlpy as dl

    project = dl.projects.get(project_name='Ocean')
    dataset = project.datasets.get(dataset_name='Sharks')

    converter = dl.Converter()
    converter.convert_dataset(dataset=dataset, to_format='voc',
                              local_path='home/voc_annotations/sharks')
